import os
import sys
import uuid
from typing import List

import inquirer
from rich.console import Console
from ruamel import yaml
from ruamel.yaml import CommentedMap, CommentedSeq

from piperider_cli import round_trip_load_yaml, safe_load_yaml, dbtutil
from piperider_cli.datasource import DATASOURCE_PROVIDERS, DataSource
from piperider_cli.error import \
    PipeRiderConfigError, \
    PipeRiderConfigTypeError, \
    PipeRiderInvalidDataSourceError, \
    DbtProjectNotFoundError, \
    DbtProfileNotFoundError

PIPERIDER_WORKSPACE_NAME = '.piperider'
PIPERIDER_WORKSPACE_PATH = os.path.join(os.getcwd(), PIPERIDER_WORKSPACE_NAME)
PIPERIDER_CONFIG_PATH = os.path.join(PIPERIDER_WORKSPACE_PATH, 'config.yml')
PIPERIDER_CREDENTIALS_PATH = os.path.join(PIPERIDER_WORKSPACE_PATH, 'credentials.yml')

# ref: https://docs.getdbt.com/dbt-cli/configure-your-profile
DBT_PROFILES_DIR_DEFAULT = '~/.dbt/'
DBT_PROFILE_FILE = 'profiles.yml'


class Configuration(object):
    """
    Configuration represents the config file in the piperider project
    at $PROJECT_ROOT/.piperider/config.yml
    """

    def __init__(self, dataSources: List[DataSource], **kwargs):
        self.dataSources: List[DataSource] = dataSources
        self.profiler_config = kwargs.get('profiler', {}) or {}
        self.includes = kwargs.get('includes', None)
        self.excludes = kwargs.get('excludes', None)
        self.include_views = kwargs.get('include_views', False)
        self.tables = kwargs.get('tables', {})
        self.telemetry_id = kwargs.get('telemetry_id', None)
        self.report_dir = self._to_report_dir(kwargs.get('report_dir', '.'))

        self._verify_input_config()
        self.includes = [str(t) for t in self.includes] if self.includes else self.includes
        self.excludes = [str(t) for t in self.excludes] if self.excludes else self.excludes

        # global dbt config
        self.dbt = kwargs.get('dbt', None)

    def _to_report_dir(self, dirname: str):
        if dirname is None or dirname.strip() == '':
            dirname = '.'
        if dirname.startswith('.'):
            return os.path.abspath(os.path.join(os.getcwd(), PIPERIDER_WORKSPACE_NAME, dirname))
        return os.path.abspath(dirname)

    def _verify_input_config(self):
        if self.profiler_config:
            limit = self.profiler_config.get('table', {}).get('limit', 0)
            if not isinstance(limit, int):
                raise PipeRiderConfigTypeError("profiler 'limit' should be an integer")

            duplicate_rows = self.profiler_config.get('table', {}).get('duplicateRows', False)
            if not isinstance(duplicate_rows, bool):
                raise PipeRiderConfigTypeError("profiler 'duplicateRows' should be an boolean")

        if self.includes is not None:
            if not isinstance(self.includes, List):
                raise PipeRiderConfigTypeError("'includes' should be a list of tables' name")

        if self.excludes is not None:
            if not isinstance(self.excludes, List):
                raise PipeRiderConfigTypeError("'excludes' should be a list of tables' name")

    def get_telemetry_id(self):
        return self.telemetry_id

    @classmethod
    def from_dbt_project(cls, dbt_project_path, dbt_profiles_dir=None):
        """
        build configuration from the existing dbt project

        :param dbt_project_path:
        :param dbt_profiles_dir:
        :return:
        """

        # Precedence reference
        # https://docs.getdbt.com/docs/get-started/connection-profiles#advanced-customizing-a-profile-directory
        if dbt_profiles_dir:
            profile_dir = dbt_profiles_dir
        elif os.getenv('DBT_PROFILES_DIR'):
            profile_dir = os.getenv('DBT_PROFILES_DIR')
        elif os.path.exists(os.path.join(os.getcwd(), DBT_PROFILE_FILE)):
            profile_dir = os.getcwd()
        else:
            profile_dir = DBT_PROFILES_DIR_DEFAULT

        dbt_profile_path = os.path.join(profile_dir, DBT_PROFILE_FILE)

        if not os.path.exists(dbt_project_path):
            raise DbtProjectNotFoundError(dbt_project_path)

        from piperider_cli.dbtutil import load_dbt_project
        dbt_project = load_dbt_project(dbt_project_path)

        if not os.path.exists(os.path.expanduser(dbt_profile_path)):
            raise DbtProfileNotFoundError(dbt_profile_path)

        dbt_profile = dbtutil.load_dbt_profile(os.path.expanduser(dbt_profile_path))

        profile_name = dbt_project.get('profile', '')
        target_name = dbt_profile.get(profile_name, {}).get('target', '')
        if target_name not in list(dbt_profile.get(profile_name, {}).get('outputs', {}).keys()):
            console = Console()
            console.print("[bold red]Error:[/bold red] "
                          f"The profile '{profile_name}' does not have a target named '{target_name}'.\n"
                          "Please check the dbt profile format.")
            sys.exit(1)
        credential = dbtutil.load_credential_from_dbt_profile(dbt_profile, profile_name, target_name)
        type_name = credential.get('type')
        dbt = {
            'projectDir': os.path.relpath(os.path.dirname(dbt_project_path), os.getcwd()),
        }

        if dbt_profiles_dir:
            dbt['profilesDir'] = dbt_profiles_dir

        if type_name not in DATASOURCE_PROVIDERS:
            raise PipeRiderInvalidDataSourceError(type_name, dbt_profile_path)

        # Set 'pass' as the alias of 'password'
        if credential.get('pass') and credential.get('password') is None:
            credential['password'] = credential.pop('pass')

        datasource_class = DATASOURCE_PROVIDERS[type_name]
        datasource: DataSource = datasource_class(name=profile_name, dbt=dbt, credential=credential)
        datasource.show_installation_information()

        return cls(dataSources=[datasource])

    @classmethod
    def load(cls, piperider_config_path=PIPERIDER_CONFIG_PATH):
        """
        load from the existing configuration

        :return:
        """
        credentials = None

        config = safe_load_yaml(piperider_config_path)
        if config is None:
            raise PipeRiderConfigError(piperider_config_path)

        data_sources: List[DataSource] = []
        ds_basic_config = ['name', 'type', 'dbt']
        for ds in config.get('dataSources', []):
            type_name = ds.get('type')
            if type_name not in DATASOURCE_PROVIDERS:
                raise PipeRiderInvalidDataSourceError(type_name, piperider_config_path)

            datasource_class = DATASOURCE_PROVIDERS[type_name]
            credential = {}
            for config_key in ds:
                if config_key in ds_basic_config:
                    continue
                credential[config_key] = ds[config_key]

            dbt = ds.get('dbt')
            if dbt:
                # Precedence reference
                # https://docs.getdbt.com/docs/get-started/connection-profiles#advanced-customizing-a-profile-directory
                if dbt.get('profilesDir'):
                    profile_dir = dbt.get('profilesDir')
                elif os.getenv('DBT_PROFILES_DIR'):
                    profile_dir = os.getenv('DBT_PROFILES_DIR')
                elif os.path.exists(os.path.join(os.getcwd(), DBT_PROFILE_FILE)):
                    profile_dir = os.getcwd()
                else:
                    profile_dir = DBT_PROFILES_DIR_DEFAULT
                profile_path = os.path.join(profile_dir, DBT_PROFILE_FILE)
                if '~' in profile_path:
                    profile_path = os.path.expanduser(profile_path)
                profile = dbtutil.load_dbt_profile(profile_path)
                profile_name = dbt.get('profile')
                target_name = dbt.get('target')
                credential.update(dbtutil.load_credential_from_dbt_profile(profile, profile_name, target_name))
                # TODO: extract duplicate code from func 'from_dbt_project'
                if credential.get('pass') and credential.get('password') is None:
                    credential['password'] = credential.pop('pass')
                data_source = datasource_class(name=ds.get('name'), dbt=dbt, credential=credential)
            else:
                try:
                    with open(PIPERIDER_CREDENTIALS_PATH, 'r') as fd:
                        credentials = yaml.safe_load(fd)
                        credential.update(credentials.get(ds.get('name'), {}))
                except FileNotFoundError:
                    pass
                except Exception:
                    raise
                data_source = datasource_class(name=ds.get('name'), credential=credential)
            data_sources.append(data_source)

        # load global dbt config
        dbt = config.get('dbt')
        if dbt:
            project_dir = config.get('dbt').get('projectDir')
            project = dbtutil.load_dbt_project(project_dir)
            profile_name = project.get('profile')

            # Precedence reference
            # https://docs.getdbt.com/docs/get-started/connection-profiles#advanced-customizing-a-profile-directory
            if dbt.get('profilesDir'):
                profile_dir = dbt.get('profilesDir')
            elif os.getenv('DBT_PROFILES_DIR'):
                profile_dir = os.getenv('DBT_PROFILES_DIR')
            elif os.path.exists(os.path.join(project_dir, DBT_PROFILE_FILE)):
                profile_dir = project_dir
            else:
                profile_dir = DBT_PROFILES_DIR_DEFAULT

            profile_path = os.path.join(profile_dir, DBT_PROFILE_FILE)
            if '~' in profile_path:
                profile_path = os.path.expanduser(profile_path)
            profile = dbtutil.load_dbt_profile(profile_path)
            if profile.get(profile_name):
                target_names = list(profile.get(profile_name).get('outputs').keys())
                for target in target_names:
                    credential = dbtutil.load_credential_from_dbt_profile(profile, profile_name, target)
                    datasource_class = DATASOURCE_PROVIDERS[credential.get('type')]
                    data_source = datasource_class(
                        name=target,
                        dbt=dict(**dbt, profile=profile_name, target=target),
                        credential=credential
                    )
                    data_sources.append(data_source)
                # dbt behavior: dbt uses 'default' as target name if no target given in profiles.yml
                dbt['target'] = profile.get(profile_name).get('target', 'default')
                if dbt['target'] not in target_names:
                    console = Console()
                    console.print("[bold red]Error:[/bold red] "
                                  f"The profile '{profile_name}' does not have a target named '{dbt['target']}'.\n"
                                  "Please check the dbt profile format.")
                    sys.exit(1)

        return cls(
            dataSources=data_sources,
            profiler=config.get('profiler', {}),
            tables=config.get('tables', {}),
            includes=config.get('includes', None),
            excludes=config.get('excludes', None),
            include_views=config.get('include_views', False),
            telemetry_id=config.get('telemetry', {}).get('id'),
            report_dir=config.get('report_dir', '.'),
            dbt=dbt
        )

    def flush_datasource(self, path):
        """
        flush the configuration to the file

        :param path:
        :return:
        """

        config = round_trip_load_yaml(path)
        if config is None:
            raise PipeRiderConfigError(path)

        def _get_exist_datasource(d, config):
            for ds in config.get('dataSources', []):
                if ds.get('name') == d.name:
                    return ds
            return None

        def _generate_datasource_config(d):
            datasource = dict(name=d.name, type=d.type_name)
            if d.args.get('dbt'):
                # dbt project
                datasource['dbt'] = d.args.get('dbt')
            else:
                # non-dbt project
                if d.credential_source == 'config':
                    datasource.update(**d.credential)
            return CommentedMap(datasource)

        flush_data_sources = CommentedSeq()

        for ds in self.dataSources:
            exist_ds = _get_exist_datasource(ds, config)
            if exist_ds:
                flush_data_sources.append(exist_ds)
            else:
                new_ds = _generate_datasource_config(ds)
                flush_data_sources.append(new_ds)
        config['dataSources'] = flush_data_sources

        if config['dataSources'][-1].ca.items == {}:
            # Only append a new line if the last datasource has no comment
            config.yaml_set_comment_before_after_key('profiler', before='\n')

        with open(path, 'w') as fd:
            yaml.YAML().dump(config, fd)
        pass

    def dump(self, path):
        """
        dump the configuration to the given path
        :param path:
        :return:
        """

        if self.telemetry_id is None:
            self.telemetry_id = uuid.uuid4().hex

        config = dict(
            dataSources=[],
            dbt=None,
            profiler=None,
            telemetry=dict(id=self.telemetry_id)
        )

        d = self.dataSources[0]
        datasource = dict(name=d.name, type=d.type_name)
        if d.args.get('dbt'):
            # dbt project
            config['dbt'] = CommentedMap(d.args.get('dbt'))
        else:
            # non-dbt project
            if d.credential_source == 'config':
                datasource.update(**d.credential)
            config.pop('dbt')
            config['dataSources'].append(datasource)

        template = '''  table:
    # the maximum row count to profile. (Default unlimited)
    limit: 1000000
    duplicateRows: false

'''

        config_yaml = CommentedMap(config)

        if config_yaml.get('dbt'):
            config_yaml.yaml_set_comment_before_after_key('profiler', before="tag: 'piperider'", indent=2)
        config_yaml.yaml_set_comment_before_after_key('profiler', before='\n')
        config_yaml.yaml_set_comment_before_after_key('telemetry', before=template)

        with open(path, 'w') as fd:
            yaml.YAML().dump(config_yaml, fd)

    def dump_credentials(self, path, after_init_config=False):
        """
        dump the credentials to the given path
        :param path:
        :param after_init_config: config file has been written
        :return:
        """
        creds = dict()
        for d in self.dataSources:
            if after_init_config and d.credential_source == 'config':
                continue
            if d.args.get('dbt'):
                continue
            creds[d.name] = dict(type=d.type_name, **d.credential)

        if creds:
            with open(path, 'w') as fd:
                yaml.round_trip_dump(creds, fd)

    def to_sqlalchemy_config(self, datasource_name):
        # TODO we will convert a data source to a sqlalchemy parameters
        raise NotImplementedError

    def ask_for_datasource(self):
        if len(self.dataSources) == 0:
            return None
        elif len(self.dataSources) == 1:
            return self.dataSources[0]
        else:
            questions = [
                inquirer.List('datasource',
                              message="Please select a datasource",
                              choices=[(f'{d.name:20} (type: {d.type_name})', d) for d in self.dataSources],
                              carousel=True)
            ]
            answers = inquirer.prompt(questions, raise_keyboard_interrupt=True)
            return answers['datasource']

    def delete_datasource(self, datasource):
        if datasource in self.dataSources:
            self.dataSources.remove(datasource)



import os
import sys
import webbrowser
from typing import List, Optional

import inquirer
import readchar
from rich import box
from rich.console import Console
from rich.prompt import Prompt
from rich.table import Table

from piperider_cli import datetime_to_str, open_report_in_browser, str_to_datetime
from piperider_cli.cloud import PipeRiderCloud
from piperider_cli.compare_report import CompareReport, RunOutput
from piperider_cli.datasource import FANCY_USER_INPUT
from piperider_cli.filesystem import FileSystem

console = Console()
piperider_cloud = PipeRiderCloud()


def ask_login_info() -> str:
    console.print('Please provide available email account to login')
    if FANCY_USER_INPUT:
        account = inquirer.text('Email address', validate=lambda _, x: '@' in x)
    else:
        while True:
            account = Prompt.ask('[[yellow]?[/yellow]] Email address')
            if '@' in account:
                break

    response = piperider_cloud.magic_login(account)
    if response is None or response.get('success') is False:
        console.print('[[red]Error[/red]] Login failed. Please try again.')
        return None

    if response.get('link'):
        webbrowser.open(response.get('link'))

    console.print('Please paste the api token from magic link. The link has been sent to your email address.')
    while True:
        api_token = Prompt.ask('[[yellow]?[/yellow]] API token')
        if len(api_token) > 0:
            break
    return api_token


def verify_api_token(api_token) -> bool:
    if piperider_cloud.validate(api_token) is False:
        console.print('[[red]Error[/red]] Invalid API Token. Please try again.')
        return False

    # Write API Token back to user profile
    piperider_cloud.service.update_api_token()
    return True


def location_mapper(loc_id):
    map = {
        'us': 'North America',
        'eu': 'Europe',
        'ap': 'Asia Pacific'
    }
    return map.get(loc_id, 'N/A')


def show_user_info():
    user = piperider_cloud.me
    if user is None:
        return
    ascii_table = Table(show_header=True, show_edge=True, header_style="bold magenta",
                        box=box.SIMPLE, title='User Profile')
    ascii_table.add_column('Email', justify='left', style='cyan')
    ascii_table.add_column('Username', justify='left')
    ascii_table.add_column('Full Name', justify='left')
    ascii_table.add_column('Storage Location', justify='left')
    ascii_table.add_column('Timezone', justify='left')
    ascii_table.add_row(
        user.get('email'),
        user.get('username', 'N/A'),
        user.get('fullname', 'N/A'),
        location_mapper(user.get('data_storage_location')),
        user.get('timezone_id')
    )
    console.print(ascii_table)
    pass


def check_default_config(options: dict):
    if options is None:
        return

    cloud_config = piperider_cloud.config

    auto_upload_flag = options.get('auto_upload')

    if cloud_config.get('auto_upload') is not None and auto_upload_flag is None:
        # Skip if auto_upload is already set
        return
    if auto_upload_flag is None and cloud_config.get('auto_upload') is None:
        console.print('Please select default behavior for auto upload')
        if FANCY_USER_INPUT:
            auto_upload_flag = inquirer.confirm('Auto upload reports to cloud', default=True)
        else:
            auto_upload_flag = Prompt.ask('[[yellow]?[/yellow]] Auto upload reports to cloud', choices=['y', 'n'],
                                          default='y') == 'y'
    console.print(f'[[bold green]Config[/bold green]] Default auto upload behavior is set to {auto_upload_flag}')
    piperider_cloud.update_config({'auto_upload': auto_upload_flag})


def select_reports(report_dir=None, datasource=None) -> List[RunOutput]:
    filesystem = FileSystem(report_dir=report_dir)
    selector = CompareReport(filesystem.get_output_dir(), None, None, datasource=datasource)
    console.rule('Select Reports to Upload')
    return selector.select_multiple_reports(action='upload')


class CloudReportOutput(RunOutput):
    def __init__(self, report):
        self.report = report
        self.id = report['id']
        self.name = report['datasource_name']
        self.created_at = report['created_at']
        self.table_count = report['tables']
        self.pass_count = report['passed']
        self.fail_count = report['failed']


def select_cloud_report_ids(project_id: int = None, datasource=None, project_name=None, target=None, base=None) -> (
    int, int):
    if target:
        target_id = get_run_report_id(target)
    if base:
        base_id = get_run_report_id(base)

    if project_id is None:
        project_id = piperider_cloud.get_default_project()

    reports = [CloudReportOutput(r) for r in piperider_cloud.list_reports(project_id, datasource=datasource)]
    if len(reports) == 0:
        return None, None

    selector = CompareReport(None, None, None, datasource=datasource, profiler_outputs=reports)

    if base is None and target is None:
        console.rule('Please select base and target reports to compare')
        base, target = selector.select_two_reports(action='compare')
        target_id = target.id
        base_id = base.id
    elif base and target is None:
        console.rule('Please select a target report to compare')
        target = selector.select_one_report(action='compare')
        target_id = target.id
    elif target and base is None:
        console.rule('Please select a base report to compare')
        base = selector.select_one_report(action='compare')
        base_id = base.id

    return base_id, target_id


def upload_to_cloud(report: RunOutput, debug=False, project_id=None) -> dict:
    response = piperider_cloud.upload_report(report.path, project_id=project_id)
    # TODO refine the output when API is ready

    if response.get('success') is True:
        project_id = response.get('project_id')
        report_id = response.get('id')
        if project_id and report_id:
            report_url = f'{piperider_cloud.service.cloud_host}/projects/{project_id}/reports/{report_id}'
        else:
            report_url = 'N/A'
        return {
            'success': True,
            'message': response.get("message"),
            'report_url': report_url,
            'file_path': report.path
        }

    if debug:
        console.print(response)
    return {
        'success': False,
        'message': response.get("message"),
        'report_url': 'N/A',
        'file_path': report.path
    }


def get_run_report_id(report_key: str) -> Optional[int]:
    if report_key.isdecimal():
        if int(report_key) < 1:
            return None
        return int(report_key)

    if report_key.startswith('datasource:'):
        datasource = report_key.split(':')[-1]
        project_id = piperider_cloud.get_default_project()
        reports = piperider_cloud.list_reports(project_id, datasource=datasource)

        if reports:
            return reports[0].get('id')

    return None


def create_compare_reports(base_id: int, target_id: int, tables_from, project_id=None) -> dict:
    if project_id is None:
        project_id = piperider_cloud.get_default_project()
    response = piperider_cloud.compare_reports(project_id, base_id, target_id, tables_from)
    if response is not None:
        url = f'{piperider_cloud.service.cloud_host}/projects/{project_id}/reports/{base_id}/comparison/{target_id}'
        console.print(f'Comparison report URL: {url}')
    else:
        console.print('Failed to create the comparison report')
    return response


class CloudConnector:
    @staticmethod
    def is_login() -> bool:
        return piperider_cloud.available

    @staticmethod
    def is_auto_upload() -> bool:
        return piperider_cloud.config.get('auto_upload', False)

    @staticmethod
    def config_auto_upload(flag: bool):
        console.print(f'[[bold green]Config[/bold green]] Default auto upload behavior is set to {flag}')
        piperider_cloud.update_config({'auto_upload': flag})

    @staticmethod
    def login(api_token=None, options: dict = None, debug=False):

        if piperider_cloud.available:
            console.rule('Already Logged In')
            show_user_info()
            check_default_config(options)
            return 0

        if api_token is None:
            api_token = ask_login_info()

        if api_token:
            if verify_api_token(api_token):
                console.rule('Login Successful')
                show_user_info()
                check_default_config(options)
                return 0

        console.rule('Login Failed', style='red')
        return 1

    @staticmethod
    def logout():

        if piperider_cloud.available is False:
            console.rule('Already Logout', style='yellow')
            return 0

        piperider_cloud.logout()
        console.rule('Logout')
        return 0

    @staticmethod
    def upload_latest_report(report_dir=None, debug=False, open_report=False, force_upload=False,
                             auto_upload=False) -> int:
        filesystem = FileSystem(report_dir=report_dir)
        latest_report_path = os.path.join(filesystem.get_output_dir(), 'latest', 'run.json')
        return CloudConnector.upload_report(latest_report_path, debug=debug, open_report=open_report,
                                            force_upload=force_upload, auto_upload=auto_upload)

    @staticmethod
    def upload_report(report_path=None, report_dir=None, datasource=None, debug=False, open_report=False,
                      force_upload=False, auto_upload=False, project_name=None) -> int:
        if piperider_cloud.available is False:
            console.rule('Please login PipeRider Cloud first', style='red')
            return 1

        project_id = piperider_cloud.get_default_project()
        if project_name is not None:
            project = piperider_cloud.get_project_by_name(project_name)
            if project is None:
                console.print(f'[[bold red]Error[/bold red]] Project \'{project_name}\' does not exist')
                return 1
            project_id = project.get('id')

        rc = 0
        results = []
        if report_path is None:
            reports = select_reports(report_dir=report_dir, datasource=datasource)
            console.rule('Uploading Reports')
            for r in reports:
                response = upload_to_cloud(r, debug, project_id=project_id)
                if response.get('success') is False:
                    rc = 1
                response['name'] = r.name
                response['created_at'] = r.created_at
                results.append(response)
        else:
            console.rule('Uploading Report')
            report = RunOutput(report_path)
            response = upload_to_cloud(report, debug, project_id=project_id)
            if response.get('success') is False:
                rc = 1
            response['name'] = report.name
            response['created_at'] = report.created_at
            results.append(response)

        console.rule('Upload Completed')
        ascii_table = Table(show_header=True, show_edge=True, header_style="bold magenta",
                            box=box.SIMPLE)
        ascii_table.add_column('Status', justify='left', style='cyan')
        ascii_table.add_column('Name', justify='left')
        ascii_table.add_column('Created At', justify='left')
        ascii_table.add_column('Report URL', justify='left', no_wrap=True)
        ascii_table.add_column('Message', justify='left')

        for response in results:
            status = '[bold green]Success[/bold green]' if response.get(
                'success') else '[bold yellow]Skipped[/bold yellow]'
            url = f"[deep_sky_blue1]{response.get('report_url', 'N/A')}[/deep_sky_blue1]"
            message = response.get('message')
            created_at = datetime_to_str(str_to_datetime(response.get('created_at')), to_tzlocal=True)
            ascii_table.add_row(status, response.get('name'), created_at, url, message)

        if open_report:
            url = response.get('report_url')
            open_report_in_browser(url, True)

        console.print(ascii_table)
        return rc

    @staticmethod
    def compare_reports(base=None, target=None, tables_from='all', summary_file=None, debug=False,
                        project_name=None) -> int:
        if piperider_cloud.available is False:
            console.rule('Please login PipeRider Cloud first', style='red')
            return 1

        project_id = None
        if project_name is not None:
            project = piperider_cloud.get_project_by_name(project_name)
            if project is None:
                console.print(f'[[bold red]Error[/bold red]] Project \'{project_name}\' does not exist')
                return 1
            project_id = project.get('id')

        base_id, target_id = select_cloud_report_ids(base=base, target=target, project_id=project_id)
        if base_id is None or target_id is None:
            raise Exception('No report found in the PipeRider Cloud. Please upload reports to PipeRider Cloud first.')

        console.print(f"Creating comparison report id={base_id} ... id={target_id}")
        response = create_compare_reports(base_id, target_id, tables_from, project_id=project_id)

        if debug:
            console.print(response)

        if summary_file:
            summary_file = os.path.abspath(summary_file)
            summary_dir = os.path.dirname(summary_file)
            if summary_dir:
                os.makedirs(summary_dir, exist_ok=True)
            with open(summary_file, 'w') as f:
                f.write(response.get('summary'))

    @staticmethod
    def list_projects(debug=False) -> int:
        if piperider_cloud.available is False:
            console.rule('Please login PipeRider Cloud first', style='red')
            return 1

        projects = piperider_cloud.list_projects()
        # console.print(projects)

        layout_table = Table(
            title="PipeRider Cloud - Project List",
            title_style='bold magenta',
            show_header=True,
            show_edge=True,
            box=box.SIMPLE_HEAVY,
        )
        layout_table.add_column('Name')
        layout_table.add_column('Type')
        layout_table.add_column('Organization')
        # layout_table.add_column('Project URL', justify='right', no_wrap=True)

        for project in projects:
            # TODO: Put the project URL in the table once the project URL is available
            # project_id = project.get('id')
            # project_url = f'[deep_sky_blue1]{piperider_cloud.service.cloud_host}/projects/{project_id}[/deep_sky_blue1]'
            project_name = project.get('name') if project.get(
                'organization_name') is None else f"{project.get('organization_name')}/{project.get('name')}"
            layout_table.add_row(
                project_name,
                project.get('parent_type'),
                project.get('organization_display_name', '-'),
                # project_url,
            )

        console.print(layout_table)
        pass

    @staticmethod
    def select_project(project_name: str = None, datasource: str = None, debug: bool = False):

        def _project_selector():
            arrow_alias_msg = ''
            if sys.platform == "win32" or sys.platform == "cygwin":
                # change readchar key UP & DOWN by 'w' and 's'
                readchar.key.UP = 'w'
                readchar.key.DOWN = 's'
                arrow_alias_msg = " 'w' to Up, 's' to Down,"

            projects = [
                (f"{p.get('organization_name')}/{p.get('name')}" if p.get('organization_name') else p.get('name'), p)
                for p in piperider_cloud.list_projects()]

            question = [
                inquirer.List('selected_project',
                              message=f"Please select a project as default project ({arrow_alias_msg} ENTER to confirm )",
                              choices=projects,
                              carousel=True,
                              )
            ]
            answers = inquirer.prompt(question, raise_keyboard_interrupt=True)
            if answers:
                return answers['selected_project']
            return None

        if project_name is None:
            project = _project_selector()
        else:
            project = piperider_cloud.get_project_by_name(project_name)
            if project is None:
                console.print(f'[[bold red]Warning[/bold red]] Project \'{project_name}\' does not exist')
                return 1

        name = project.get('name') if project.get(
            'organization_name') is None else f"{project.get('organization_name')}/{project.get('name')}"
        piperider_cloud.set_default_project(name)

        # TODO: Add project name into the datasource config if datasource is not None
        console.print(f'[[bold green]Config[/bold green]] Default project is set to \'{name}\'')
        pass



from piperider_cli.assertion_engine import AssertionContext, AssertionResult
from piperider_cli.assertion_engine.assertion import ValidationResult
from piperider_cli.assertion_engine.types.base import BaseAssertionType

COLUMN_TYPES = ['string', 'integer', 'numeric', 'datetime', 'date', 'time', 'boolean', 'other']


class AssertColumnSchemaType(BaseAssertionType):
    def name(self):
        return "assert_column_schema_type"

    def execute(self, context: AssertionContext):
        return assert_column_schema_type(context)

    def validate(self, context: AssertionContext) -> ValidationResult:
        # type list: https://docs.sqlalchemy.org/en/14/core/type_basics.html#sql-standard-and-multiple-vendor-types
        return ValidationResult(context).require('schema_type', str)


class AssertColumnType(BaseAssertionType):
    def name(self):
        return "assert_column_type"

    def execute(self, context: AssertionContext):
        return assert_column_type(context)

    def validate(self, context: AssertionContext) -> ValidationResult:
        result = ValidationResult(context).require('type', str)
        if result.has_errors():
            return result

        if not set([context.asserts.get("type")]).issubset(set(COLUMN_TYPES)):
            result.errors.append(
                f'type parameter should be one of {COLUMN_TYPES}, input: {context.asserts.get("type")}')
        return result


class AssertColumnInTypes(BaseAssertionType):
    def name(self):
        return "assert_column_in_types"

    def execute(self, context: AssertionContext):
        return assert_column_in_types(context)

    def validate(self, context: AssertionContext) -> ValidationResult:
        result = ValidationResult(context).require('types', list)
        if result.has_errors():
            return result

        if not set(context.asserts.get("types")).issubset(set(COLUMN_TYPES)):
            result.errors.append(f'types parameter should be one of {COLUMN_TYPES}, '
                                 f'input: {context.asserts.get("types")}')

        return result


def assert_column_schema_type(context: AssertionContext) -> AssertionResult:
    table = context.table
    column = context.column
    metrics = context.profiler_result

    column_metrics = metrics.get('tables', {}).get(table, {}).get('columns', {}).get(column)
    if not column_metrics:
        return context.result.fail_with_metric_not_found_error(context.table, context.column)

    # Check assertion input
    assert_schema_type = context.asserts.get('schema_type').upper()
    if not assert_schema_type:
        return context.result.fail_with_assertion_error('Expect a SQL schema type')
    context.result.expected = assert_schema_type

    schema_type = column_metrics.get('schema_type')
    context.result.actual = schema_type

    if schema_type == assert_schema_type:
        return context.result.success()

    return context.result.fail()


def assert_column_type(context: AssertionContext) -> AssertionResult:
    table = context.table
    column = context.column
    metrics = context.profiler_result

    column_metrics = metrics.get('tables', {}).get(table, {}).get('columns', {}).get(column)
    if not column_metrics:
        return context.result.fail_with_metric_not_found_error(context.table, context.column)

    # Check assertion input
    assert_type = context.asserts.get('type').lower()
    if not assert_type:
        return context.result.fail_with_assertion_error(f'Expect a type in {COLUMN_TYPES}')

    if assert_type not in COLUMN_TYPES:
        return context.result.fail_with_assertion_error(f'The column type should one of {COLUMN_TYPES}.')

    context.result.expected = assert_type

    column_type = column_metrics.get('type')

    context.result.actual = column_type

    if column_type == assert_type:
        return context.result.success()

    return context.result.fail()


def assert_column_in_types(context: AssertionContext) -> AssertionResult:
    table = context.table
    column = context.column
    metrics = context.profiler_result

    column_metrics = metrics.get('tables', {}).get(table, {}).get('columns', {}).get(column)
    if not column_metrics:
        return context.result.fail_with_metric_not_found_error(context.table, context.column)

    # Check assertion input
    assert_types = context.asserts.get('types', [])
    assert_types = [x.lower() for x in assert_types]
    if not assert_types:
        return context.result.fail_with_assertion_error(f'Expect a list of types in {COLUMN_TYPES}')

    invalid_types = [t for t in assert_types if t not in COLUMN_TYPES]
    if invalid_types:
        return context.result.fail_with_assertion_error(
            f'Invalid types {invalid_types}. The column type should one of {COLUMN_TYPES}.')

    context.result.expected = assert_types

    column_type = column_metrics.get('type')

    context.result.actual = column_type

    if column_type in set(assert_types):
        return context.result.success()

    return context.result.fail()



import io
import json
import os
import shutil
import sys
from datetime import datetime, date
from typing import List

import inquirer
import readchar
from rich.console import Console

import piperider_cli.hack.inquirer as inquirer_hack
from piperider_cli import datetime_to_str, str_to_datetime, clone_directory, \
    raise_exception_when_directory_not_writable
from piperider_cli.filesystem import FileSystem
from piperider_cli.generate_report import setup_report_variables


class RunOutput(object):
    def __init__(self, path):
        self.path = path
        self.name = None
        self.created_at = None

        self.table_count = 0
        self.pass_count = 0
        self.fail_count = 0

        try:
            with open(path, 'r') as f:
                run_result = json.load(f)
                self.name = run_result['datasource']['name']
                self.created_at = run_result['created_at']

                tables = run_result.get('tables', {})
                self.table_count = len(tables.keys())

                for test in run_result.get('tests', []):
                    if test.get('status') == 'passed':
                        self.pass_count += 1
                    else:
                        self.fail_count += 1
        except Exception as e:
            if isinstance(e, json.decoder.JSONDecodeError):
                raise json.decoder.JSONDecodeError(
                    f'Invalid JSON in file "{path}"', e.doc, e.pos)
            raise e

    def verify(self) -> bool:
        # TODO: add some verification logic
        return True

    def load(self):
        with open(self.path, 'r') as f:
            data = json.load(f)
        return data

    def __str__(self):
        created_at_str = datetime_to_str(str_to_datetime(self.created_at),
                                         to_tzlocal=True)

        return f'{self.name:12} ' \
               f'#table={self.table_count:<6} ' \
               f'#pass={self.pass_count:<5} ' \
               f'#fail={self.fail_count:<5} ' \
               f'{created_at_str}'


def _merge_keys(base: List[str], target: List[str]):
    '''
    Merge keys from base, target tables. Unlike default union, it preserves the order for column rename, added, removed.

    :param base: keys for base table
    :param target: keys for base table
    :return: merged keys
    '''

    result = []
    while base and target:
        if base[0] == target[0]:
            result.append(base[0])
            base.pop(0)
            target.pop(0)
        elif base[0] in target:
            idx = target.index(base[0])
            for i in target[0:idx]:
                if i not in result:
                    result.append(i)
            result.append(base[0])
            base.pop(0)
            target = target[idx + 1:]
        else:
            result.append(base[0])
            base.pop(0)

    for c in base:
        if c not in result:
            result.append(c)

    for c in target:
        if c not in result:
            result.append(c)

    return result


def join(base, target):
    '''
    Join base and target to a dict which

    keys = (base keys) +  (target keys)
    result[keys] = {base: {...}, target: {...}

    :param base:
    :param target:
    :return:
    '''
    if not base:
        base = dict()
    if not target:
        target = dict()

    keys = _merge_keys(list(base.keys()), list(target.keys()))
    result = dict()
    for key in keys:
        value = dict()
        value['base'] = base.get(key)
        value['target'] = target.get(key)
        result[key] = value
    return result


class ComparisonData(object):
    STATE_ADD = 0
    STATE_DEL = 1
    STATE_MOD = 2

    def __init__(self, base, target, tables_from):
        self._id = datetime.now().strftime("%Y%m%d%H%M%S")

        if tables_from == 'target-only':
            target_run_tables = target.get('tables', {}).keys()
            target_run_metrics = {metric.get('name') for metric in target.get('metrics', [])}
            target_run_tests = {test.get('id') for test in target.get('tests', [])}

            base['tables'] = {table_name: base['tables'][table_name] for table_name in base['tables'] if
                              table_name in target_run_tables}
            base['metrics'] = [metric for metric in base['metrics'] if metric.get('name') in target_run_metrics]
            base['tests'] = [test for test in base['tests'] if test.get('id') in target_run_tests]
        elif tables_from == 'base-only':
            base_run_tables = base.get('tables', {}).keys()
            base_run_metrics = {metric.get('name') for metric in base.get('metrics', [])}
            base_run_tests = {test.get('id') for test in base.get('tests', [])}

            target['tables'] = {table_name: target['tables'][table_name] for table_name in target['tables'] if
                                table_name in base_run_tables}
            target['metrics'] = [metric for metric in target['metrics'] if metric.get('name') in base_run_metrics]
            target['tests'] = [test for test in target['tests'] if test.get('id') in base_run_tests]

        self._base = base
        self._target = target

    def id(self):
        return self._id

    def to_json(self):
        output = dict(
            created_at=datetime_to_str(datetime.utcnow()),
            base=self._base,
            # TODO: rename input -> target in schema and result json
            input=self._target,
        )
        return json.dumps(output, separators=(',', ':'))

    def to_summary_markdown(self):
        out = io.StringIO()

        # Comparison summary
        out.write(self._render_compare_markdown())

        # Per-table Comparison
        base = self._base.get('tables')
        target = self._target.get('tables')

        states = {}
        per_table_out = io.StringIO()
        joined = join(base, target)
        for table_name in joined.keys():
            joined_table = joined[table_name]

            columns_b = joined_table.get('base').get('columns') if joined_table.get('base') else None
            columns_t = joined_table.get('target').get('columns') if joined_table.get('target') else None

            state, result = self._render_table_summary_markdown(table_name, columns_b, columns_t)
            if state == self.STATE_ADD:
                if states.get('added') is None:
                    states['added'] = 0
                states['added'] += 1
            elif state == self.STATE_DEL:
                if states.get('deleted') is None:
                    states['deleted'] = 0
                states['deleted'] += 1
            elif state == self.STATE_MOD:
                if states.get('schema changed') is None:
                    states['schema changed'] = 0
                states['schema changed'] += 1
            per_table_out.write(result)

        # Per-table summary
        states = sorted(states.items())
        table_summary_hint = ', '.join([f'{state}={num}' for state, num in states])
        if table_summary_hint:
            table_summary_hint = f' ({table_summary_hint})'

        out.write("<details>\n")
        out.write(f"<summary>Tables Summary{table_summary_hint}</summary>\n")
        out.write("<blockquote>\n")
        out.write("\n")
        out.write(per_table_out.getvalue())
        out.write("</blockquote></details>")

        base = self._base.get('metrics', [])
        target = self._target.get('metrics', [])
        out.write(self._render_metrics_comparison_markdown(base, target))

        return out.getvalue()

    @staticmethod
    def _value_with_annotation(key, annotation=None):
        annotation_str = f" ({annotation})" if annotation else ''
        return f"{key}{annotation_str}"

    @staticmethod
    def _value_with_change(base, target):
        if target is None:
            return f"~~{base}~~"
        elif base is None:
            return target
        elif base != target:
            return f"~~{base}~~<br/>{target}"
        else:
            return target

    @staticmethod
    def _value_with_delta(base, target, percentage=False):
        if target is None:
            return '-'

        annotation = '%' if percentage else ''
        delta = ''
        if base is not None:
            diff = target - base
            if percentage:
                diff *= 100
            delta = f" ({'+' if target >= base else ''}{round(diff, 2)}{annotation})"

        if percentage:
            target = round(target * 100, 2)

        return f"{target}{annotation}{delta}"

    @staticmethod
    def _display_metrics_delta(b_value, t_value):
        if b_value == '' or t_value == '':
            return ''
        elif isinstance(b_value, str) and b_value.startswith('~~'):
            return ''
        elif isinstance(t_value, str) and t_value.startswith('~~'):
            return ''
        elif b_value == t_value:
            return '-'
        b_value = 0 if b_value == '-' else b_value
        t_value = 0 if t_value == '-' else t_value
        delta = t_value - b_value

        return f"{'+' if delta >= 0 else ''}{delta}"

    @staticmethod
    def _get_column_changed(base, target):
        columns_b = base.get('columns') if base else None
        columns_t = target.get('columns') if target else None
        joined = join(columns_b, columns_t)
        added = 0
        deleted = 0
        changed = 0

        for column_name in joined.keys():
            joined_column = joined[column_name]
            b = joined_column.get('base')
            t = joined_column.get('target')

            schema_type_b = b.get('schema_type') if b else None
            schema_type_t = t.get('schema_type') if t else None

            if b is None:
                added += 1
            elif t is None:
                deleted += 1
            elif schema_type_b != schema_type_t:
                changed += 1

        return {
            'added': added,
            'deleted': deleted,
            'changed': changed,
        }

    def _render_compare_markdown(self):
        base = self._base.get('tables')
        target = self._target.get('tables')

        out = io.StringIO()
        joined = join(base, target)
        print("Table | Rows | Columns ", file=out)
        print("--- | --- | ---", file=out)
        for table_name in joined.keys():
            joined_table = joined[table_name]
            b = joined_table.get('base')
            t = joined_table.get('target')

            annotation = None
            change_message_str = ''

            if b is None:
                annotation = '+'
            elif t is None:
                annotation = '-'
            else:
                column_changed = self._get_column_changed(b, t)
                messages = []

                if column_changed['added'] > 0:
                    annotation = '!'
                    messages.append(f"added={column_changed['added']}")
                if column_changed['deleted'] > 0:
                    annotation = '!'
                    messages.append(f"deleted={column_changed['deleted']}")
                if column_changed['changed'] > 0:
                    annotation = '!'
                    messages.append(f"changed={column_changed['changed']}")

                if messages:
                    change_message_str = ', '.join(messages)
                    change_message_str = f"<br/>({change_message_str})"

            rows_b, cols_b = (b.get('row_count', 0), b.get('col_count', 0)) if b else (None, None)
            rows_t, cols_t = (t.get('row_count', 0), t.get('col_count', 0)) if t else (None, None)

            out.write(f"{self._value_with_annotation(table_name, annotation)} | ")
            out.write(f"{self._value_with_delta(rows_b, rows_t)} | ")
            out.write(f"{self._value_with_delta(cols_b, cols_t)}{change_message_str} \n")

        base = self._base.get('metrics', [])
        target = self._target.get('metrics', [])
        base_metrics = {base_metric.get('name'): base_metric for base_metric in base}
        target_metrics = {target_metric.get('name'): target_metric for target_metric in target}

        joined = join(base_metrics, target_metrics)
        out_metrics = io.StringIO()
        if joined:
            def _display_period_annotation(period_grain):
                if period_grain == 'day':
                    return '(Yesterday)'
                else:
                    return f'(Last {period_grain})'

            print("Metric | Period | Base | Target | +/- ", file=out_metrics)
            print("--- | --- | :-: | :-: | :-: ", file=out_metrics)
            for metric_name in joined.keys():
                joined_metric = joined[metric_name]
                b = joined_metric.get('base')
                t = joined_metric.get('target')
                b_data = {}
                t_data = {}
                if b:
                    grain = b.get('grain')
                    metric_label = b.get('label')
                    b_data = {row[0]: row[1] for row in b.get('data')}
                if t:
                    grain = t.get('grain')
                    metric_label = t.get('label')
                    t_data = {row[0]: row[1] for row in t.get('data')}

                t_dates = t_data.keys() if t_data.keys() else b_data.keys()
                t_dates = sorted(list(t_dates), key=lambda k: date.fromisoformat(k), reverse=True)
                t_last_date = t_dates[1]

                if t_last_date not in b_data:
                    b_result = ''
                else:
                    b_result = b_data.get(t_last_date) if b_data.get(t_last_date) is not None else '-'

                if t_last_date not in t_data:
                    t_result = ''
                else:
                    t_result = t_data.get(t_last_date) if t_data.get(t_last_date) is not None else '-'

                out_metrics.write(f"{metric_label} | {t_last_date} {_display_period_annotation(grain)} | ")
                out_metrics.write(f"{b_result} | {t_result} | {self._display_metrics_delta(b_result, t_result)} \n ")

        return f"""<details>
<summary>Comparison Summary</summary>

{out.getvalue()}
{out_metrics.getvalue()}
</details>
"""

    def _render_table_summary_markdown(self, table_name, base, target):
        out = io.StringIO()
        joined = join(base, target)
        print("Column | Type | Valid % | Distinct %", file=out)
        print("--- | --- | --- | ---", file=out)
        table_modified = False
        for column_name in joined.keys():
            joined_column = joined[column_name]
            b = joined_column.get('base')
            t = joined_column.get('target')

            schema_type_b = self._get_metric_from_report(b, 'schema_type', None)
            valids_p_b = self._get_metric_from_report(b, 'valids_p', None)
            distinct_p_b = self._get_metric_from_report(b, 'distinct_p', None)

            schema_type_t = self._get_metric_from_report(t, 'schema_type', None)
            valids_p_t = self._get_metric_from_report(t, 'valids_p', None)
            distinct_p_t = self._get_metric_from_report(t, 'distinct_p', None)

            annotation = None
            if b is None:
                annotation = '+'
                table_modified = True
            elif t is None:
                annotation = '-'
                table_modified = True
            elif schema_type_b != schema_type_t:
                annotation = '!'
                table_modified = True

            out.write(f"{self._value_with_annotation(column_name, annotation)} | ")
            out.write(f"{self._value_with_change(schema_type_b, schema_type_t)} | ")
            out.write(f"{self._value_with_delta(valids_p_b, valids_p_t, percentage=True)} | ")
            out.write(f"{self._value_with_delta(distinct_p_b, distinct_p_t, percentage=True)} \n")

        annotation = ''
        state = None
        if base is None:
            annotation = ' (+)'
            state = self.STATE_ADD
        elif target is None:
            annotation = ' (-)'
            state = self.STATE_DEL
        elif table_modified:
            annotation = ' (!)'
            state = self.STATE_MOD

        return state, f"""<details>
<summary>{table_name}{annotation}</summary>

{out.getvalue()}
</details>
"""

    @staticmethod
    def _get_metric_from_report(report, metric, default):
        return report.get(metric, default) if report else default

    def _render_metrics_comparison_markdown(self, base, target):
        out = io.StringIO()

        base_metrics = {base_metric.get('name'): base_metric for base_metric in base}
        target_metrics = {target_metric.get('name'): target_metric for target_metric in target}
        joined = join(base_metrics, target_metrics)
        if joined:
            out_metrics = io.StringIO()
            states = {}
            for metric_name in joined.keys():
                notation = ''
                metric_warn = False
                joined_metric = joined[metric_name]
                b = joined_metric.get('base')
                t = joined_metric.get('target')
                date_grain = 'date'
                b_data = {}
                t_data = {}
                if b:
                    date_grain = b.get('headers')[0]
                    date_grain = date_grain[0].upper() + date_grain[1:]
                    metric_label = b.get('label')
                    b_data = {row[0]: row[1] for row in b.get('data')}
                else:
                    notation = '(+)'
                    if states.get('added') is None:
                        states['added'] = 0
                    states['added'] += 1
                if t:
                    date_grain = t.get('headers')[0]
                    date_grain = date_grain[0].upper() + date_grain[1:]
                    metric_label = t.get('label')
                    t_data = {row[0]: row[1] for row in t.get('data')}
                else:
                    notation = '(-)'
                    if states.get('deleted') is None:
                        states['deleted'] = 0
                    states['deleted'] += 1

                b_dates = b_data.keys() if b_data.keys() else t_data.keys()
                b_dates = sorted(list(b_dates), key=lambda k: date.fromisoformat(k), reverse=True)
                b_latest_date = b_dates[0]
                t_dates = t_data.keys() if t_data.keys() else b_data.keys()
                t_dates = sorted(list(t_dates), key=lambda k: date.fromisoformat(k), reverse=True)
                t_latest_date = t_dates[0]

                out_metric = io.StringIO()
                for d in t_dates:
                    if d not in b_data:
                        b_result = ''
                    else:
                        b_result = b_data.get(d) if b_data.get(d) is not None else '-'
                        if date.fromisoformat(d) == date.fromisoformat(b_latest_date):
                            b_result = f'~~{b_data.get(d)}~~' if b_data.get(d) is not None else '-'

                    if d not in t_data:
                        t_result = ''
                    else:
                        t_result = t_data.get(d) if t_data.get(d) is not None else '-'
                        if date.fromisoformat(d) == date.fromisoformat(t_latest_date):
                            t_result = f'~~{t_data.get(d)}~~' if t_data.get(d) is not None else '-'

                    if date.fromisoformat(t_latest_date) >= date.fromisoformat(d) >= date.fromisoformat(b_latest_date):
                        delta = ''
                    else:
                        delta = self._display_metrics_delta(b_result, t_result)
                        if delta != '-' and delta != '':
                            metric_warn = True
                    out_metric.write(f"{d} | {b_result} | {t_result} | {delta}\n")

                if metric_warn:
                    notation = '(!)'
                    if states.get('metric changed') is None:
                        states['metric changed'] = 0
                    states['metric changed'] += 1

                out_metrics.write("<details>\n")
                out_metrics.write(f"<summary>{metric_label} {notation}</summary>\n\n")
                if not t:
                    out_metrics.write("Metric is not available in target\n")
                else:
                    out_metrics.write(f"{date_grain} | Base | Target | -/+ \n")
                    out_metrics.write(":-: | :-: | :-: | :-: \n")
                    out_metrics.write(out_metric.getvalue())
                out_metrics.write("</details>\n")

            states = sorted(states.items())
            metrics_summary_hint = ', '.join([f'{state}={num}' for state, num in states])
            if metrics_summary_hint:
                metrics_summary_hint = f' ({metrics_summary_hint})'
            out.write("<details>\n")
            out.write(f"<summary>Metrics Summary{metrics_summary_hint}</summary>\n")
            out.write("<blockquote>\n\n")
            out.write(out_metrics.getvalue())
            out.write("</blockquote></details>")

        return out.getvalue()


def prepare_default_output_path(filesystem: FileSystem, created_at):
    latest_symlink_path = os.path.join(filesystem.get_comparison_dir(), 'latest')
    latest_source = created_at
    comparison_path = os.path.join(filesystem.get_comparison_dir(), created_at)

    if not os.path.exists(comparison_path):
        os.makedirs(comparison_path, exist_ok=True)

    # Create a symlink pointing to the latest comparison directory
    if os.path.islink(latest_symlink_path):
        os.unlink(latest_symlink_path)
    if not os.path.exists(latest_symlink_path):
        os.symlink(latest_source, latest_symlink_path)
    else:
        console = Console()
        console.print(f'[bold yellow]Warning: {latest_symlink_path} already exists[/bold yellow]')

    return comparison_path


class CompareReport(object):
    def __init__(self, profiler_output_path, a=None, b=None, datasource=None, profiler_outputs=None):
        self.profiler_output_path = profiler_output_path
        self.profiler_outputs = profiler_outputs
        self.console = Console()
        self.a: RunOutput = RunOutput(a) if a else None
        self.b: RunOutput = RunOutput(b) if b else None
        self.datasource = datasource

    def list_existing_outputs(self, output_search_path=None) -> List[RunOutput]:
        """
        List existing profiler outputs.
        """

        def _walk_throw_runs(path):
            outputs = []
            for root, dirs, _ in os.walk(path):
                for dir in dirs:
                    if dir == 'latest':
                        continue
                    run_json = os.path.join(root, dir, 'run.json')
                    if not os.path.exists(run_json):
                        continue
                    output = RunOutput(run_json)
                    if self.datasource and output.name != self.datasource:
                        continue
                    outputs.append(output)
            outputs.sort(key=lambda x: (x.name, x.created_at), reverse=True)
            return outputs

        if output_search_path is None:
            output_search_path = self.profiler_output_path

        return _walk_throw_runs(output_search_path)

    def get_the_last_two_reports(self):
        outputs = self.list_existing_outputs()
        outputs.sort(key=lambda x: x.created_at)
        outputs = list(filter(lambda x: x.name == outputs[-1].name, outputs))
        if len(outputs) < 2:
            return None, None
        return outputs[-2:]

    def select_reports(self, use_last_two=None):
        if use_last_two:
            self.a, self.b = self.get_the_last_two_reports()
        elif self.a is None and self.b is None:
            self.a, self.b = self.select_two_reports()
        elif self.a and self.b is None:
            self.b = self.select_one_report()
            pass
        elif self.a is None and self.b:
            self.a = self.select_one_report()

        if self.a and self.b:
            self.console.print('Selected reports:')
            self.console.print(f'  Base:   {self.a.path}')
            self.console.print(f'  Target: {self.b.path}')
            return True
        return False

    def select_multiple_reports(self, action='compare', limit=None):
        def _report_validater(answers, current) -> bool:
            if limit is None:
                return len(current) > 0
            else:
                return len(current) == limit

        profiler_outputs = self.profiler_outputs if self.profiler_outputs is not None else self.list_existing_outputs()
        arrow_alias_msg = ''
        if sys.platform == "win32" or sys.platform == "cygwin":
            # change readchar key UP & DOWN by 'w' and 's'
            readchar.key.UP = 'w'
            readchar.key.DOWN = 's'
            arrow_alias_msg = " 'w' to Up, 's' to Down,"

        if limit is None:
            if len(profiler_outputs) == 0:
                raise Exception("Not enough reports to compare. Please run 'piperider run' first.")
            questions = [
                inquirer.Checkbox('profiler_output',
                                  message=f"Please select the reports to {action} ({arrow_alias_msg} SPACE to select, and ENTER to confirm )",
                                  choices=profiler_outputs,
                                  carousel=True,
                                  validate=_report_validater,
                                  )
            ]
        else:
            if len(profiler_outputs) < limit:
                raise Exception("Not enough reports to compare. Please run 'piperider run' first.")
            report_msg = 'a report' if limit == 1 else f'the {limit} reports'
            questions = [
                inquirer_hack.LimitedCheckboxQuestion('profiler_output',
                                                      message=f"Please select {report_msg} to {action} ({arrow_alias_msg} SPACE to select, and ENTER to confirm )",
                                                      choices=profiler_outputs,
                                                      carousel=True,
                                                      validate=_report_validater,
                                                      limited=limit,
                                                      )
            ]

        answers = inquirer_hack.prompt_ex(questions, raise_keyboard_interrupt=True)
        if answers:
            return answers['profiler_output']
        return None

    def select_one_report(self, action='compare') -> RunOutput:
        answers = self.select_multiple_reports(action=action, limit=1)
        if answers:
            return answers[0]
        else:
            return None

    def select_two_reports(self, action='compare') -> (RunOutput, RunOutput):
        """
        Select multiple files from a list of files.
        """

        answers = self.select_multiple_reports(action=action, limit=2)
        if answers:
            return answers[0], answers[1]
        return None, None

    def generate_data(self, tables_from: str) -> ComparisonData:
        if self.a is None or self.b is None:
            raise Exception("Please select reports to compare first.")

        return ComparisonData(self.a.load(), self.b.load(), tables_from)

    @staticmethod
    def exec(*, a=None, b=None, last=None, datasource=None, report_dir=None, output=None, tables_from='all',
             summary_file=None, debug=False):
        console = Console()

        filesystem = FileSystem(report_dir=report_dir)
        raise_exception_when_directory_not_writable(output)

        report = CompareReport(filesystem.get_output_dir(), a, b, datasource=datasource)
        if not report.select_reports(use_last_two=last):
            raise Exception('No valid reports found')
        comparison_data = report.generate_data(tables_from)

        from piperider_cli import data
        report_template_dir = os.path.join(os.path.dirname(data.__file__), 'report', 'comparison-report')
        with open(os.path.join(report_template_dir, 'index.html')) as f:
            report_template_html = f.read()

        def output_report(directory):
            clone_directory(report_template_dir, directory)
            filename = os.path.join(directory, 'index.html')
            with open(filename, 'w') as f:
                html = setup_report_variables(report_template_html, False, comparison_data.to_json())
                f.write(html)

        def output_summary(directory):
            filename = os.path.join(directory, 'summary.md')
            with open(filename, 'w') as f:
                f.write(comparison_data.to_summary_markdown())

        data_id = comparison_data.id()
        default_report_directory = prepare_default_output_path(filesystem, data_id)
        output_report(default_report_directory)
        output_summary(default_report_directory)
        report_path = os.path.join(filesystem.get_comparison_dir(), 'latest', 'index.html')
        summary_md_path = os.path.join(filesystem.get_comparison_dir(), 'latest', 'summary.md')

        if output:
            clone_directory(default_report_directory, output)
            report_path = os.path.abspath(os.path.join(output, 'index.html'))
            summary_md_path = os.path.abspath(os.path.join(output, 'summary.md'))

        if summary_file:
            summary_file = os.path.abspath(summary_file)
            summary_dir = os.path.dirname(summary_file)
            if summary_dir:
                os.makedirs(summary_dir, exist_ok=True)
            shutil.copyfile(summary_md_path, summary_file)
            summary_md_path = summary_file

        console.print()
        console.print(f"Comparison report: {report_path}")
        console.print(f"Comparison summary: {summary_md_path}")

        if debug:
            # Write comparison data to file
            with open(os.path.join(default_report_directory, 'comparison_data.json'), 'w') as f:
                f.write(comparison_data.to_json())



import asyncio
import concurrent.futures
import decimal
import itertools
from concurrent.futures import ThreadPoolExecutor
from typing import List, Union

from sqlalchemy import select, func, distinct, literal_column, join, outerjoin, Column, Date
from sqlalchemy.engine import Engine
from sqlalchemy.pool import SingletonThreadPool
from sqlalchemy.sql.expression import text, union_all, case
from sqlalchemy.sql.selectable import CTE

from piperider_cli.datasource import DataSource
from piperider_cli.metrics_engine.event import MetricEventHandler, DefaultMetricEventHandler


def dtof(value: Union[int, float, decimal.Decimal]) -> Union[int, float]:
    """
    dtof is helpler function to transform decimal value to float. Decimal is not json serializable type.

    :param value:
    :return:
    """
    if isinstance(value, decimal.Decimal):
        return float(value)
    return value


class Metric:
    def __init__(
        self,
        name,
        table,
        schema,
        database,
        expression,
        timestamp,
        calculation_method,
        time_grains=None,
        dimensions=None,
        filters=None,
        label=None,
        description=None
    ):
        self.name = name
        self.table = table
        self.database = database
        self.schema = schema.lower() if schema is not None else None
        self.expression = expression
        self.timestamp = timestamp
        self.calculation_method = calculation_method
        self.time_grains = time_grains
        self.dimensions = dimensions
        self.filters = filters
        self.label = label
        self.description = description
        self.ref_metrics: List[Metric] = []


class MetricEngine:
    """
    Profiler profile tables and columns by a sqlalchemy engine.
    """

    def __init__(
        self,
        data_source: DataSource,
        metrics,
        event_handler: MetricEventHandler = DefaultMetricEventHandler()
    ):
        self.data_source = data_source
        self.metrics = metrics
        self.event_handler = event_handler
        if self.data_source.threads > 1:
            self.executor = ThreadPoolExecutor(max_workers=self.data_source.threads)
        else:
            self.executor = None

    @staticmethod
    def _get_query_param(metric: Metric) -> (str, List[str]):
        for grain in metric.time_grains:
            if grain not in ['day', 'week', 'month', 'quarter', 'year']:
                continue
            if not metric.dimensions:
                yield grain, []
            else:
                for r in range(1, len(metric.dimensions) + 1):
                    for dims in itertools.combinations(metric.dimensions, r):
                        yield grain, list(dims)

    @staticmethod
    def _compose_query_name(grain: str, dimensions: List[str], label=False) -> str:
        if grain == 'day':
            grain = 'daily'
        else:
            grain += 'ly'

        if label:
            grain = grain.upper()[0] + grain[1:]

        return grain

    def _get_query_stmt(self, metric: Metric, grain: str, dimension: List[str], date_spine_model: CTE):
        metric_column_name = metric.name

        if metric.calculation_method == 'derived':
            selectable = None

            # Join all parent metrics
            for ref_metric in metric.ref_metrics:
                cte = self._get_query_stmt(ref_metric, grain, dimension, date_spine_model).cte(
                    f'{ref_metric.name}_model')
                if selectable is None:
                    selectable = cte
                else:
                    selectable = join(selectable, cte, selectable.c.d == cte.c.d)

            # a / b / c -> a / nullif(b, 0) / nullif(c, 0)
            expression = metric.expression
            if '/' in expression:
                expression_list = expression.split('/')
                dividend = expression_list[0]
                divisors = [f'nullif({divisor}, 0)' for divisor in expression_list[1:]]
                expression = f"{dividend} / {'/'.join(divisors)}"

            return select([
                cte.c.d,
                literal_column(expression).label(metric_column_name)
            ]).select_from(
                selectable
            ).order_by(
                cte.c.d,
            )
        else:
            # Source model
            if self.data_source.type_name == 'bigquery':
                source_model = text(f"`{metric.database}.{metric.schema}.{metric.table}`")
            else:
                source_model = text(f"{metric.database}.{metric.schema}.{metric.table}")

            # Base model
            # 1. map expression to 'c'
            # 2. map truncated timestamp to 'd'
            # 3. Filter only the last n grain
            # 4. Filter according to metric filters
            start_date = self.date_trunc(grain, func.current_date()) - self._interval(grain,
                                                                                      self._slot_count_by_grain(grain))
            stmt = select([
                literal_column(metric.expression).label('c'),
                self.date_trunc(grain, func.cast(literal_column(metric.timestamp), Date)).label('d'),
            ]).select_from(
                source_model
            ).where(
                func.cast(literal_column(metric.timestamp), Date) >= start_date
            )
            for f in metric.filters:
                stmt = stmt.where(text(f"{f.get('field')} {f.get('operator')} {f.get('value')}"))
            base_model = stmt.cte(f"{metric.name}_base_model")

            # Aggregated model
            # 1. select 'd'
            # 1. aggrgate by calculation method as 'm'
            # 1. group by 'd'

            if metric.calculation_method == 'count':
                agg_expression = func.count(base_model.columns['c'])
            elif metric.calculation_method == 'count_distinct':
                agg_expression = func.count(distinct(base_model.columns['c']))
            elif metric.calculation_method == 'sum':
                agg_expression = func.sum(base_model.columns['c'])
            elif metric.calculation_method == 'average':
                agg_expression = func.avg(base_model.columns['c'])
            elif metric.calculation_method == 'min':
                agg_expression = func.min(base_model.columns['c'])
            elif metric.calculation_method == 'max':
                agg_expression = func.max(base_model.columns['c'])
            else:
                return None

            agg_model = select([
                base_model.c.d,
                agg_expression.label('m')
            ]).select_from(
                base_model
            ).group_by(
                base_model.c.d
            ).cte(name=f"{metric.name}_agg_model")

            metric_column = agg_model.c.m
            if metric.calculation_method in ['count', 'count_distinct', 'sum']:
                metric_column = case(
                    [(metric_column.is_(None), 0)],
                    else_=metric_column
                )

            return select([
                date_spine_model.c.d,
                metric_column.label(metric_column_name)
            ]).select_from(
                outerjoin(date_spine_model, agg_model, date_spine_model.c.d == agg_model.c.d)
            ).order_by(
                date_spine_model.c.d
            )

    def _interval(self, grain: str, n):
        if self.data_source.type_name == 'bigquery':
            return text(f"interval {n} {grain}")
        else:
            if grain == 'quarter':
                return text(f"interval '{n * 3} months'")
            else:
                return text(f"interval '{n} {grain}s'")

    def _slot_count_by_grain(self, grain: str):
        if grain == 'year':
            return 10
        elif grain == 'quarter':
            return 12
        elif grain == 'month':
            return 12
        elif grain == 'week':
            return 12
        else:
            return 30

    def _date_spine(self, grain: str):
        '''
        Generate the date spine for left join
        https://github.com/dbt-labs/dbt-utils#date_spine-source
        '''

        n = self._slot_count_by_grain(grain)
        dates = []
        for i in range(n):
            one_date = select([
                (self.date_trunc(grain, func.current_date()) - self._interval(grain, i + 1)).label('d')
            ])
            dates.append(one_date)

        current_date = select([
            self.date_trunc(grain, func.current_date()).label('d')
        ])

        return union_all(*dates, current_date)

    def _query_metric(self, engine: Engine, metric: Metric, grain: str, dimension: List[str]) -> dict:
        '''
        Query a metric with given parameter. Just implement the behavior of 'metrics.calculate'
        ref: https://docs.getdbt.com/docs/build/metrics#querying-your-metric

        :param metric:
        :param grain:
        :param dimension:
        :return:
        '''
        headers = [f'date_{grain}'] + dimension + [metric.name]

        query_result = {
            'name': f'{metric.name}_{self._compose_query_name(grain, dimension)}',
            'label': f'{metric.label} ({self._compose_query_name(grain, dimension, label=True)})',
            'description': metric.description,
            'grain': grain,
            'dimensions': dimension,
            'headers': headers,
            'data': []
        }

        with engine.connect() as conn:
            date_spine_model = self._date_spine(grain).cte(name="date_spine_model")
            stmt = self._get_query_stmt(metric, grain, dimension, date_spine_model)

            result = conn.execute(stmt)

            for row in result:
                row = list(row)
                row[0] = str(row[0])
                row[-1] = dtof(row[-1])
                query_result['data'].append(row)

        return self._compose_query_name(grain, dimension), query_result

    async def _execute(self) -> List[dict]:
        metrics = self.metrics
        results = []

        self.event_handler.handle_run_start()
        total_metric = len(metrics)
        completed_metric = 0

        for metric in metrics:
            self.event_handler.handle_run_progress(total_metric, completed_metric)
            self.event_handler.handle_metric_start(metric.label)

            total_param = len(list(self._get_query_param(metric)))
            completed_param = 0
            engine = self.data_source.get_engine_by_database(metric.database)

            query_results = {}
            futures = []
            loop = asyncio.get_running_loop()
            for grain, dimension in self._get_query_param(metric):
                # to keep the order
                query_param = self._compose_query_name(grain, dimension)
                query_results[query_param] = None

                future = loop.run_in_executor(self.executor, self._query_metric, engine, metric, grain, dimension)
                futures.append(future)

            for future in asyncio.as_completed(futures):
                query_param, result = await future
                self.event_handler.handle_metric_progress(metric.label, total_param, completed_param)
                self.event_handler.handle_param_query_start(metric.label, query_param)
                query_results[query_param] = result
                self.event_handler.handle_param_query_end(metric.label)
                completed_param += 1

            for query_result in query_results.values():
                results.append(query_result)

            self.event_handler.handle_metric_end(metric.label)
            completed_metric += 1
        self.event_handler.handle_run_end()
        return results

    def execute(self) -> List[dict]:
        if self.executor:
            with self.executor:
                return asyncio.run(self._execute())
        else:
            return asyncio.run(self._execute())

    def date_trunc(self, date_part, date_expression) -> Column:
        type_name = self.data_source.type_name
        if type_name == 'sqlite':
            if date_part == "YEAR":
                return func.strftime("%Y-01-01", date_expression)
            elif date_part == "MONTH":
                return func.strftime("%Y-%m-01", date_expression)
            else:
                return func.strftime("%Y-%m-%d", date_expression)
        elif type_name == 'bigquery':
            return func.date_trunc(date_expression, text(date_part))
        else:
            return func.date_trunc(date_part, date_expression)



import os.path
import sys

import click
import sentry_sdk
from rich.console import Console

import piperider_cli.dbtutil as dbtutil
from piperider_cli import __version__, sentry_dns, sentry_env, event
from piperider_cli.assertion_generator import AssertionGenerator
from piperider_cli.cloud_connector import CloudConnector
from piperider_cli.compare_report import CompareReport
from piperider_cli.configuration import PIPERIDER_WORKSPACE_PATH
from piperider_cli.event import UserProfileConfigurator
from piperider_cli.event.track import TrackCommand
from piperider_cli.exitcode import EC_ERR_TEST_FAILED
from piperider_cli.feedback import Feedback
from piperider_cli.generate_report import GenerateReport
from piperider_cli.guide import Guide
from piperider_cli.initializer import Initializer
from piperider_cli.recipe_executor import RecipeExecutor
from piperider_cli.runner import Runner
from piperider_cli.validator import Validator

release_version = __version__ if sentry_env != 'development' else None

sentry_sdk.init(
    sentry_dns,
    environment=sentry_env,
    release=release_version,
    # Set traces_sample_rate to 1.0 to capture 100%
    # of transactions for performance monitoring.
    # We recommend adjusting this value in production.
    traces_sample_rate=1.0
)
sentry_sdk.set_tag("piperider.version", __version__)
sentry_sdk.set_tag("platform", sys.platform)

event.init()

debug_option = [
    click.option('--debug', is_flag=True, help='Enable debug mode.')
]


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


@click.group(name="piperider", invoke_without_command=True)
@click.pass_context
def cli(ctx: click.Context):
    'An open-source toolkit for detecting data issues across pipelines that works with CI systems for continuous data quality assessment.'
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        Guide().show_tips(ctx.command.name)


# It works after click>=8
cli.command_class = TrackCommand


@cli.command(short_help='Show version information.', cls=TrackCommand)
def version():
    'Show version information.'
    console = Console()
    console.print(__version__)

    if 'production' == sentry_env:
        return

    from piperider_cli import data
    commit_file = os.path.abspath(os.path.join(os.path.dirname(data.__file__), 'COMMIT'))
    try:
        with open(commit_file) as fh:
            commit_sha = fh.read().strip()
            console.print(f'GitCommit: {commit_sha}')
    except Exception:
        pass


@cli.command(short_help='Send your feedback to help us improve the PipeRider.', cls=TrackCommand)
def feedback():
    'Send your feedback to help us improve the PipeRider.'
    Feedback.exec()


@cli.command(short_help='Initialize a PipeRider project.', cls=TrackCommand)
@click.option('--no-auto-search', type=click.BOOL, default=False, is_flag=True,
              help="Disable auto detection of dbt projects.")
@click.option('--dbt-project-dir', type=click.Path(exists=True), default=None,
              help='Directory to search for dbt_project.yml.')
@click.option('--dbt-profiles-dir', type=click.Path(exists=True), default=None,
              help='Directory to search for dbt profiles.yml.')
@add_options(debug_option)
def init(**kwargs):
    'Initialize a PipeRider project in interactive mode. The configurations are saved in ".piperider".'

    console = Console()

    # TODO show the process and message to users
    console.print(f'Initialize piperider to path {PIPERIDER_WORKSPACE_PATH}')

    # Search dbt project config files
    dbt_project_path = None
    dbt_project_dir = kwargs.get('dbt_project_dir')
    dbt_profiles_dir = kwargs.get('dbt_profiles_dir')
    no_auto_search = kwargs.get('no_auto_search')
    if not no_auto_search:
        if dbt_project_dir:
            dbt_project_path = os.path.join(dbt_project_dir, "dbt_project.yml")
        else:
            dbt_project_path = dbtutil.search_dbt_project_path()

    if dbt_project_path:
        console.print(f'[[bold green] DBT [/bold green]] Use the existing dbt project file: {dbt_project_path}')

    config = Initializer.exec(dbt_project_path=dbt_project_path, dbt_profiles_dir=dbt_profiles_dir)
    if kwargs.get('debug'):
        for ds in config.dataSources:
            console.rule('Configuration')
            console.print(ds.__dict__)
    if config is None:
        sys.exit(1)

    # Show the content of config.yml
    Initializer.show_config()


@cli.command(short_help='Check project configuration.', cls=TrackCommand)
@add_options(debug_option)
def diagnose(**kwargs):
    'Check project configuration, datasource, connections, and assertion configuration.'

    console = Console()
    console.print('Diagnosing...')

    console.print(f'[bold dark_orange]PipeRider Version:[/bold dark_orange] {__version__}')

    if not Validator.diagnose():
        sys.exit(1)


@cli.command(short_help='Profile data source, run assertions, and generate report(s).', cls=TrackCommand)
@click.option('--datasource', default=None, type=click.STRING, help='Datasource to use.', metavar='DATASOURCE_NAME')
@click.option('--table', default=None, type=click.STRING, help='Table to use.', metavar='TABLE_NAME')
@click.option('--output', '-o', default=None, type=click.STRING, help='Directory to save the results.')
@click.option('--skip-report', is_flag=True, help='Skip generating report.')
@click.option('--dbt-state', default=None, help='Directory of the the dbt state.')
@click.option('--dbt-list', is_flag=True, help='Associate with dbt list format input.')
@click.option('--dbt-run-results', is_flag=True, help='Associate with dbt run results.')
@click.option('--report-dir', default=None, type=click.STRING, help='Use a different report directory.')
@click.option('--upload', is_flag=True, help='Upload the report to the PipeRider Cloud.')
@click.option('--open', is_flag=True, help='Opens the generated report in the system\'s default browser')
@add_options(debug_option)
def run(**kwargs):
    'Profile data source, run assertions, and generate report(s). By default, the raw results and reports are saved in ".piperider/outputs".'

    datasource = kwargs.get('datasource')
    table = kwargs.get('table')
    output = kwargs.get('output')
    open_report = kwargs.get('open')
    skip_report = kwargs.get('skip_report')
    dbt_state_dir = kwargs.get('dbt_state')
    dbt_list = kwargs.get('dbt_list')
    dbt_run_results = kwargs.get('dbt_run_results')
    force_upload = kwargs.get('upload')

    console = Console()

    # True -> 1, False -> 0
    if sum([True if table else False, dbt_list, dbt_run_results]) > 1:
        console.print("[bold red]Error:[/bold red] "
                      "['--table', '--dbt-list', '--dbt-run-results'] are mutually exclusive")
        sys.exit(1)

    dbt_resources = None
    if dbt_list:
        metrics = []
        models = []
        for dbt_resource in sys.stdin:
            if dbt_resource.startswith('source:'):
                continue
            elif dbt_resource.startswith('metric:'):
                metrics.append(dbt_resource.rstrip().replace('metric:', 'metric.'))
            else:
                models.append(dbt_resource.rstrip())
        dbt_resources = dict(metrics=metrics, models=models)

    ret = Runner.exec(datasource=datasource,
                      table=table,
                      output=output,
                      skip_report=skip_report,
                      dbt_state_dir=dbt_state_dir,
                      dbt_run_results=dbt_run_results,
                      dbt_resources=dbt_resources,
                      report_dir=kwargs.get('report_dir'))
    if ret in (0, EC_ERR_TEST_FAILED):
        auto_upload = CloudConnector.is_auto_upload()
        is_cloud_view = (force_upload or auto_upload)

        if CloudConnector.is_login() and is_cloud_view:
            CloudConnector.upload_latest_report(report_dir=kwargs.get('report_dir'), debug=kwargs.get('debug'),
                                                open_report=open_report, force_upload=force_upload,
                                                auto_upload=auto_upload)

        if not skip_report:
            GenerateReport.exec(None, kwargs.get('report_dir'), output, open_report, is_cloud_view)
    if ret != 0:
        sys.exit(ret)
    return ret


@cli.command(short_help='Generate recommended assertions.', cls=TrackCommand)
@click.option('--input', default=None, type=click.Path(exists=True), help='Specify the raw result file.')
@click.option('--no-recommend', is_flag=True, help='Generate assertions templates only.')
@click.option('--report-dir', default=None, type=click.STRING, help='Use a different report directory.')
@click.option('--table', default=None, type=click.STRING, help='Generate assertions for the given table')
@add_options(debug_option)
def generate_assertions(**kwargs):
    'Generate recommended assertions based on the latest result. By default, the profiling result will be loaded from ".piperider/outputs".'
    input_path = kwargs.get('input')
    report_dir = kwargs.get('report_dir')
    no_recommend = kwargs.get('no_recommend')
    table = kwargs.get('table')
    ret = AssertionGenerator.exec(input_path=input_path, report_dir=report_dir, no_recommend=no_recommend, table=table)
    if ret != 0:
        sys.exit(ret)
    return ret


@cli.command(short_help='Generate a report.', cls=TrackCommand)
@click.option('--input', default=None, type=click.Path(exists=True), help='Specify the raw result file.')
@click.option('--output', '-o', default=None, type=click.STRING, help='Directory to save the results.')
@click.option('--report-dir', default=None, type=click.STRING, help='Use a different report directory.')
@add_options(debug_option)
def generate_report(**kwargs):
    'Generate a report from the latest raw result or specified result. By default, the raw results are saved in ".piperider/outputs".'
    GenerateReport.exec(input=kwargs.get('input'), report_dir=kwargs.get('report_dir'), output=kwargs.get('output'))


@cli.command(short_help='Compare two existing reports.', cls=TrackCommand)
@click.option('--base', default=None, type=click.Path(exists=True), help='Specify the base report file.')
@click.option('--target', default=None, type=click.Path(exists=True), help='Specify the report file to be compared.')
@click.option('--last', default=None, is_flag=True, help='Compare the last two reports.')
@click.option('--datasource', default=None, type=click.STRING, metavar='DATASOURCE_NAME',
              help='Specify the datasource.')
@click.option('--output', '-o', default=None, type=click.STRING, help='Directory to save the results.')
@click.option('--report-dir', default=None, type=click.STRING, help='Use a different report directory.')
@click.option('--tables-from', default='all',
              type=click.Choice(['all', 'target-only', 'base-only'], case_sensitive=False),
              help='Show table comparison from base or target.')
@add_options(debug_option)
def compare_reports(**kwargs):
    'Compare two existing reports selected in interactive mode or by option.'

    a = kwargs.get('base')
    b = kwargs.get('target')
    last = kwargs.get('last')
    datasource = kwargs.get('datasource')
    tables_from = kwargs.get('tables_from')
    CompareReport.exec(a=a, b=b, last=last, datasource=datasource,
                       report_dir=kwargs.get('report_dir'), output=kwargs.get('output'), tables_from=tables_from,
                       debug=kwargs.get('debug', False))


@cli.group('config', short_help='Manage the PipeRider configurations.')
def config(**kwargs):
    pass


@config.command(name='list-datasource', short_help='List DataSources in current PipeRider project',
                cls=TrackCommand)
@add_options(debug_option)
def list(**kwargs):
    'List PipeRider current configurations.'
    Initializer.list()


@config.command(name='add-datasource', short_help='Add a DataSource to current PipeRider project.',
                cls=TrackCommand)
@add_options(debug_option)
def add(**kwargs):
    Initializer.add()
    pass


@config.command(name='delete-datasource', short_help='Delete a DataSource from current PipeRider project.',
                cls=TrackCommand)
@add_options(debug_option)
def delete(**kwargs):
    Initializer.list()
    Initializer.delete()
    pass


@config.command(name='enable-auto-upload', short_help='Enable auto upload to PipeRider Cloud.', cls=TrackCommand)
def enable_auto_upload(**kwargs):
    CloudConnector.config_auto_upload(True)
    pass


@config.command(name='disable-auto-upload', short_help='Disable auto upload to PipeRider Cloud.', cls=TrackCommand)
def disable_auto_upload(**kwargs):
    CloudConnector.config_auto_upload(False)
    pass


@config.command(name='enable-user-tracking', short_help='Enable user tracking.', cls=TrackCommand)
def enable_user_tracking(**kwargs):
    UserProfileConfigurator.update('anonymous_tracking', True, name='user tracking')


@config.command(name='disable-user-tracking', short_help='Disable user tracking.', cls=TrackCommand)
def disable_user_tracking(**kwargs):
    UserProfileConfigurator.update('anonymous_tracking', False, name='user tracking')


@cli.group('cloud', short_help='Manage PipeRider Cloud')
def cloud(**kwargs):
    # Manage PipeRider Cloud.
    pass


@cloud.command(short_help='Upload a report to the PipeRider Cloud.', cls=TrackCommand)
@click.option('--run', type=click.Path(exists=True), help='Specify the raw result file.')
@click.option('--report-dir', default=None, type=click.STRING, help='Use a different report directory.')
@click.option('--datasource', default=None, type=click.STRING, metavar='DATASOURCE_NAME',
              help='Specify the datasource.')
@click.option('--project', default=None, type=click.STRING, metavar='PROJECT_NAME',
              help='Specify the project to upload.')
@add_options(debug_option)
def upload_report(**kwargs):
    """
    Upload a single run report to PipeRider Cloud
    """
    report_path = kwargs.get('run')
    datasource = kwargs.get('datasource')
    report_dir = kwargs.get('report_dir')
    project_name = kwargs.get('project')
    ret = CloudConnector.upload_report(report_path=report_path, datasource=datasource, report_dir=report_dir,
                                       project_name=project_name,
                                       debug=kwargs.get('debug', False))
    return ret


@cloud.command(name='compare-reports', short_help='Generate comparison report on PipeRider Cloud.', cls=TrackCommand)
@click.option('--base', default=None, required=False, type=click.STRING,
              help='Specify the base report id or data source name. e.g., 123 or datasource:<name>')
@click.option('--target', default=None, required=False, type=click.STRING,
              help='Specify the target report id or data source name. e.g., 123 or datasource:<name>')
@click.option('--tables-from', default='all',
              type=click.Choice(['all', 'target-only', 'base-only'], case_sensitive=False),
              help='Show table comparison from base or target.')
@click.option('--summary-file', default=None, type=click.STRING, help='Download the comparison summary markdown file.')
@click.option('--project', default=None, type=click.STRING, metavar='PROJECT_NAME',
              help='Specify the project to upload.')
@add_options(debug_option)
def cloud_compare_reports(**kwargs):
    """
    Generate comparison report on PipeRider Cloud
    """
    base = kwargs.get('base')
    target = kwargs.get('target')
    tables_from = kwargs.get('tables_from')
    summary_file = kwargs.get('summary_file')
    project_name = kwargs.get('project')

    ret = CloudConnector.compare_reports(base=base, target=target, tables_from=tables_from, summary_file=summary_file,
                                         project_name=project_name, debug=kwargs.get('debug', False))

    if ret != 0:
        sys.exit(ret)
    return ret


@cli.command(name='compare', short_help='Generate comparison report with the recipe.', cls=TrackCommand)
@click.option('--recipe', default=None, type=click.STRING, help='Select a different recipe.')
@click.option('--output', '-o', default=None, type=click.STRING, help='Directory to save the results.')
@click.option('--summary-file', default=None, type=click.STRING, help='Output the comparison summary markdown file.')
@add_options(debug_option)
def compare_with_recipe(**kwargs):
    """
    Generate comparison report with the recipe
    """

    recipe = kwargs.get('recipe')
    summary_file = kwargs.get('summary_file')

    ret = 0
    try:
        RecipeExecutor.exec(recipe_name=recipe)
        CompareReport.exec(a=None, b=None, last=True, datasource=None,
                           output=kwargs.get('output'), tables_from="all",
                           summary_file=summary_file,
                           debug=kwargs.get('debug', False))
    except Exception as e:
        raise e

    return ret


@cloud.command(short_help='Login to PipeRider Cloud.', cls=TrackCommand)
@click.option('--token', default=None, type=click.STRING, help='Specify the API token.')
@click.option('--enable-auto-upload', default=None, is_flag=True, help='Enable auto upload.')
@click.option('--disable-auto-upload', default=None, is_flag=True, help='Disable auto upload.')
@add_options(debug_option)
def login(**kwargs):
    options = {
        # True for enable, False for disable, None for not defined
        'auto_upload': None,
    }
    if kwargs.get('enable_auto_upload'):
        options['auto_upload'] = True
    elif kwargs.get('disable_auto_upload'):
        options['auto_upload'] = False

    ret = CloudConnector.login(api_token=kwargs.get('token'), options=options, debug=kwargs.get('debug', False))
    return ret


@cloud.command(short_help='Logout from PipeRider Cloud.', cls=TrackCommand)
@add_options(debug_option)
def logout(**kwargs):
    ret = CloudConnector.logout()
    return ret


@cloud.command(short_help='List projects on PipeRider Cloud.', cls=TrackCommand)
@add_options(debug_option)
def list_projects(**kwargs):
    ret = CloudConnector.list_projects()
    return ret


@cloud.command(short_help='Select a project on PipeRider Cloud as default project.', cls=TrackCommand)
@click.option('--project', default=None, type=click.STRING, metavar='PROJECT_NAME',
              help='Specify the project name.')
@add_options(debug_option)
def select_project(**kwargs):
    project_name = kwargs.get('project')
    ret = CloudConnector.select_project(project_name=project_name)
    return ret



import inquirer.errors as errors
import inquirer.themes as themes
from inquirer.questions import Question
from inquirer.render.console import ConsoleRender
from inquirer.render.console._checkbox import Checkbox
from inquirer.render.console._confirm import Confirm
from inquirer.render.console._editor import Editor
from inquirer.render.console._list import List
from inquirer.render.console._password import Password
from inquirer.render.console._path import Path
from inquirer.render.console._text import Text
from readchar import key


class HackedConsoleRender(ConsoleRender):
    def render_factory(self, question_type):
        matrix = {
            "text": Text,
            "editor": Editor,
            "password": Password,
            "confirm": Confirm,
            "list": List,
            "checkbox": Checkbox,
            "limited_checkbox": LimitedCheckbox,
            "path": Path,
        }

        if question_type not in matrix:
            raise errors.UnknownQuestionTypeError()
        return matrix.get(question_type)


class LimitedCheckboxQuestion(Question):
    kind = 'limited_checkbox'

    def __init__(self, name, message="", choices=None, default=None, ignore=False, validate=True, carousel=False,
                 limited=0):
        super().__init__(name, message, choices, default, ignore, validate)
        self.carousel = carousel
        if limited > 0 and limited <= len(choices):
            self.limited = limited


class LimitedCheckbox(Checkbox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.limited = self.question.limited
        if self.limited > 0 and self.limited <= len(self.selection):
            self.selection = self.selection[-self.limited:]

    def process_input(self, pressed):
        question = self.question
        if pressed == key.UP:
            if question.carousel and self.current == 0:
                self.current = len(question.choices) - 1
            else:
                self.current = max(0, self.current - 1)
            return
        elif pressed == key.DOWN:
            if question.carousel and self.current == len(question.choices) - 1:
                self.current = 0
            else:
                self.current = min(len(self.question.choices) - 1, self.current + 1)
            return
        elif pressed == key.SPACE:
            if self.current in self.selection:
                self.selection.remove(self.current)
            else:
                if question.limited != 0 and len(self.selection) == question.limited:
                    self.selection.pop(0)
                self.selection.append(self.current)
        elif pressed == key.LEFT:
            if self.current in self.selection:
                self.selection.remove(self.current)
        elif pressed == key.RIGHT:
            if self.current not in self.selection:
                self.selection.append(self.current)
        elif pressed == key.ENTER:
            result = []
            for x in self.selection:
                value = self.question.choices[x]
                result.append(getattr(value, "value", value))
            raise errors.EndOfInput(result)
        elif pressed == key.CTRL_C:
            raise KeyboardInterrupt()


def prompt_ex(questions, render=None, answers=None, theme=themes.Default(), raise_keyboard_interrupt=False):
    render = render or HackedConsoleRender(theme=theme)
    answers = answers or {}

    try:
        for question in questions:
            answers[question.name] = render.render(question, answers)
        return answers
    except KeyboardInterrupt:
        if raise_keyboard_interrupt:
            raise
        print("")
        print("Cancelled by user")
        print("")



import json
from typing import Optional


class RecommendedAssertion:
    def __init__(self, name: Optional[str], metric: Optional[str], asserts: dict):
        self.table: str = None
        self.column: str = None
        self.name: str = name
        self.metric: str = metric
        self.asserts: dict = asserts

    def __repr__(self):
        table = self.table if self.table else ''
        column = ('.' + self.column) if self.column else ''

        asserts = json.dumps(self.asserts).replace('\"', '') if self.asserts else ''

        return f'{table}{column}: {self.name} {asserts}'



import os.path

from rich.console import Console

from piperider_cli.configuration import Configuration
from piperider_cli.recipes import select_recipe_file, RecipeConfiguration, execute_configuration, DEFAULT_RECIPE_PATH
from piperider_cli.recipes.default_recipe_generator import generate_default_recipe, show_recipe_content

console = Console()


class RecipeExecutor():
    @staticmethod
    def exec(recipe_name: str, auto_generate_default_recipe: bool = True):
        recipe_path = select_recipe_file(recipe_name)

        if recipe_path is None:
            if auto_generate_default_recipe:
                config = Configuration.load()
                if config.dataSources and config.dataSources[0].args.get('dbt'):
                    dbt_project_path = os.path.relpath(config.dataSources[0].args.get('dbt', {}).get('projectDir'))
                # generate a default recipe
                console.rule("Recipe executor: generate default recipe")
                generate_default_recipe(dbt_project_path=dbt_project_path)
                recipe_path = DEFAULT_RECIPE_PATH
                show_recipe_content(recipe_path)
            else:
                raise FileNotFoundError(f"Cannot find the recipe '{recipe_name}'")
        cfg = RecipeConfiguration.load(recipe_path)
        execute_configuration(cfg)



import json
import os
import sys
from glob import glob
from typing import Optional

import inquirer
from rich.console import Console
from rich.table import Table
from ruamel import yaml

from piperider_cli.error import \
    DbtProjectInvalidError, \
    DbtProfileInvalidError, \
    DbtProfileBigQueryAuthWithTokenUnsupportedError
from piperider_cli.metrics_engine import Metric

console = Console()


def search_dbt_project_path():
    exclude_patterns = ['site-packages', 'dbt_packages']
    _warning_if_search_path_too_widely(os.getcwd())
    paths = glob(os.path.join(os.getcwd(), '**', 'dbt_project.yml'), recursive=True)
    for exclude_pattern in exclude_patterns:
        paths = list(filter(lambda x: exclude_pattern not in x, paths))
    dbt_project_path = None
    if not paths:
        # No dbt project found
        return dbt_project_path

    if len(paths) == 1:
        # Only one dbt project found, use it
        dbt_project_path = paths[0]
    else:
        # Multiple dbt projects found, ask user to select one
        paths = sorted(paths, key=len)
        default_index = 0
        console.print(f'\nMultiple dbt projects found. Please select the dbt project: (Default: {default_index})')
        table = Table(show_header=True, header_style="bold magenta")
        table.add_column('Index', style="bold", width=5, justify='right')
        table.add_column('Path', style="bold")
        for i, p in enumerate(paths):
            table.add_row(str(i), p)
        console.print(table)
        in_index = input('index : ').strip()
        try:
            in_index = int(in_index) if in_index else default_index
            index = int(in_index)
            dbt_project_path = paths[index]
        except Exception as e:
            console.print(f'[bold yellow]Failed to read index![/] Error: {e}')
            return None

    return dbt_project_path


def _warning_if_search_path_too_widely(search_path):
    # Only warning user on macOS platform
    if sys.platform != "darwin":
        return

    home_dir = os.path.expanduser('~')

    if search_path in home_dir:
        console.print(
            f"[[bold yellow]Warning[/bold yellow]] Search path '{search_path}' is too widely. It will take some time to parse directories and may need extra permissions.")
        if inquirer.confirm(message="Do you still want to keep going?", default=True) is not True:
            raise KeyboardInterrupt()


def is_ready(config):
    if not config.get('profile'):
        return False

    if not config.get('target'):
        return False

    if not config.get('projectDir'):
        return False

    if not os.path.expanduser(config.get('projectDir')):
        return False

    return True


def _is_dbt_file_existing(dbt_state_dir: str, filename: str):
    dbt_state_dir = os.path.abspath(dbt_state_dir)
    filepath = os.path.join(dbt_state_dir, filename)
    if not os.path.exists(filepath):
        return False
    return True


def is_dbt_state_ready(dbt_state_dir: str):
    return _is_dbt_file_existing(dbt_state_dir, 'manifest.json')


def is_dbt_run_results_ready(dbt_state_dir: str):
    return _is_dbt_file_existing(dbt_state_dir, 'run_results.json')


def _get_state_run_results(dbt_state_dir: str):
    path = os.path.join(dbt_state_dir, 'run_results.json')
    with open(path) as f:
        run_results = json.load(f)

    return run_results


def _get_state_manifest(dbt_state_dir: str):
    path = os.path.join(dbt_state_dir, 'manifest.json')
    with open(path) as f:
        manifest = json.load(f)

    return manifest


def append_descriptions(profile_result, dbt_state_dir):
    manifest = _get_state_manifest(dbt_state_dir)

    nodes = manifest.get('nodes')
    for node in nodes.values():
        model = node.get('name')
        model_desc = node.get('description')
        if model not in profile_result['tables']:
            continue
        if model_desc:
            profile_result['tables'][model]['description'] = f"{model_desc} - via DBT"

        columns = node.get('columns', {})
        for column, v in columns.items():
            if column not in profile_result['tables'][model]['columns']:
                continue
            column_desc = v.get('description')
            if column_desc:
                profile_result['tables'][model]['columns'][column]['description'] = f"{column_desc} - via DBT"


def get_dbt_state_candidate(dbt_state_dir: str, options: dict):
    candidate = []
    material_whitelist = ['seed', 'table', 'incremental']
    resource_whitelist = ['model']
    if options.get('view_profile'):
        material_whitelist.append('view')

    tag = options.get('tag')
    dbt_run_results = options.get('dbt_run_results')
    dbt_resources = options.get('dbt_resources')

    manifest = _get_state_manifest(dbt_state_dir)
    nodes = manifest.get('nodes')

    run_results_ids = []
    if dbt_run_results:
        run_results = _get_state_run_results(dbt_state_dir)
        for result in run_results.get('results'):
            if result.get('status') != 'success':
                continue
            run_results_ids.append(result.get('unique_id'))

    def is_chosen(key, node):
        if dbt_resources:
            return '.'.join(node.get('fqn')) in dbt_resources['models']
        else:
            if dbt_run_results and key not in run_results_ids:
                return False
            if tag:
                return tag in node.get('tags', [])
            config_material = node.get('config').get('materialized')
            if config_material not in material_whitelist:
                return False
            return True

    for key, node in nodes.items():
        if node.get('resource_type') not in resource_whitelist:
            continue
        if not is_chosen(key, node):
            continue
        candidate.append(node)

    return candidate


def get_dbt_state_tests_result(dbt_state_dir: str):
    output = []
    unique_tests = {}

    run_results = _get_state_run_results(dbt_state_dir)
    manifest = _get_state_manifest(dbt_state_dir)

    nodes = manifest.get('nodes')
    sources = manifest.get('sources')
    for result in run_results.get('results', []):
        unique_id = result.get('unique_id')

        node = nodes.get(unique_id)
        if node.get('resource_type') != 'test':
            continue

        unique_tests[unique_id] = dict(
            status=result.get('status'),
            failures=result.get('failures'),
            message=result.get('message'),
        )

        test_node = node
        table = None
        depends_on_nodes = test_node.get('depends_on', {}).get('nodes', [])
        for depends_on_node_id in depends_on_nodes:
            depends_on_node = nodes.get(depends_on_node_id)
            if depends_on_node_id.startswith('source'):
                source = sources.get(depends_on_node_id)
                table = f"{source.get('source_name')}.{source.get('name')}"
                break
            elif depends_on_node.get('resource_type') in ['model', 'seed']:
                table = depends_on_node.get('name')
                break
        column = test_node.get('column_name')

        if table is None:
            continue

        output.append(dict(
            id=unique_id,
            name=unique_id,
            table=table,
            column=column if column != test_node['name'] else None,
            status='failed' if unique_tests[unique_id]['status'] == 'fail' else 'passed',
            tags=[],
            message=unique_tests[unique_id]['message'],
            display_name=test_node['name'],
            source='dbt'
        ))

    return output


def get_dbt_state_metrics(dbt_state_dir: str, dbt_tag: str, dbt_resources: Optional[dict] = None):
    manifest = _get_state_manifest(dbt_state_dir)

    def is_chosen(key, metric):
        if dbt_resources:
            return key in dbt_resources['metrics']
        return dbt_tag in metric.get('tags')

    metrics = []
    metric_map = {}
    for key, metric in manifest.get('metrics').items():
        if metric.get('calculation_method') == 'derived':
            table = None
            schema = None
            database = None
        else:
            depends_on_node = metric.get('depends_on').get('nodes')[0]
            table = manifest.get('nodes').get(depends_on_node).get('alias')
            schema = manifest.get('nodes').get(depends_on_node).get('schema')
            database = manifest.get('nodes').get(depends_on_node).get('database')

        m = Metric(metric.get('name'), table, schema, database, metric.get('expression'), metric.get('timestamp'),
                   metric.get('calculation_method'), metric.get('time_grains'), dimensions=None,
                   filters=metric.get('filters'), label=metric.get('label'), description=metric.get('description'))

        metric_map[key] = m

        if is_chosen(key, metric):
            if metric.get('window'):
                console.print(
                    f"[[bold yellow]Warning[/bold yellow]] Skip metric '{metric.get('name')}'. Property 'window' is not supported.")
                continue
            metrics.append(m)

    for key, metric in metric_map.items():
        if metric.calculation_method == 'derived':
            for depends_on_metric in manifest.get('metrics').get(key).get('depends_on').get('nodes'):
                metric.ref_metrics.append(metric_map.get(depends_on_metric))

    return metrics


def load_dbt_project(path: str):
    if not path.endswith('dbt_project.yml'):
        path = os.path.join(path, 'dbt_project.yml')

    with open(path, 'r') as fd:
        try:
            yml = yaml.YAML()
            yml.allow_duplicate_keys = True
            return yml.load(fd)
        except Exception as e:
            raise DbtProjectInvalidError(path, e)


def load_dbt_profile(path):
    from jinja2 import Environment, FileSystemLoader

    def env_var(var, default=None):
        return os.getenv(var, default)

    def as_bool(var):
        return var.lower() in ('true', 'yes', '1')

    def as_number(var):
        if var.isnumeric():
            return int(var)
        return float(var)

    def as_text(var):
        return str(var)

    env = Environment(loader=FileSystemLoader(searchpath=os.path.dirname(path)))
    env.globals['env_var'] = env_var
    env.filters['as_bool'] = as_bool
    env.filters['as_number'] = as_number
    env.filters['as_text'] = as_text
    template = env.get_template(os.path.basename(path))
    try:
        yml = yaml.YAML()
        yml.allow_duplicate_keys = True
        return yml.load(template.render())
    except Exception as e:
        raise DbtProfileInvalidError(path, e)


def load_credential_from_dbt_profile(dbt_profile, profile_name, target_name):
    credential = dbt_profile.get(profile_name, {}).get('outputs', {}).get(target_name, {})

    if credential.get('type') == 'bigquery':
        # BigQuery Data Source
        from piperider_cli.datasource.bigquery import AUTH_METHOD_OAUTH_SECRETS
        # DBT profile support 4 types of methods to authenticate with BigQuery:
        #   [ 'oauth', 'oauth-secrets', 'service-account', 'service-account-json' ]
        # Ref: https://docs.getdbt.com/reference/warehouse-profiles/bigquery-profile#authentication-methods
        if credential.get('method') == 'oauth-secrets':
            credential['method'] = AUTH_METHOD_OAUTH_SECRETS
            # TODO: Currently SqlAlchemy haven't support using access token to authenticate with BigQuery.
            #       Ref: https://github.com/googleapis/python-bigquery-sqlalchemy/pull/459
            raise DbtProfileBigQueryAuthWithTokenUnsupportedError
    elif credential.get('type') == 'redshift':
        if credential.get('method') is None:
            credential['method'] = 'password'
        host = credential.get('host')
        port = credential.get('port')
        dbname = credential.get('dbname')
        credential['endpoint'] = f'{host}:{port}/{dbname}'
    return credential



import os

from piperider_cli import raise_exception_when_directory_not_writable
from piperider_cli.configuration import PIPERIDER_WORKSPACE_NAME, Configuration

piperider_default_report_dir = os.path.join(os.getcwd(), PIPERIDER_WORKSPACE_NAME)


class FileSystem:

    def __init__(self, **kwargs):
        config = Configuration.load()
        self.report_dir = config.report_dir
        if kwargs.get('report_dir', None) is not None:
            self.report_dir = self._to_report_dir(kwargs.get('report_dir'))

        raise_exception_when_directory_not_writable(self.report_dir)

    def _to_report_dir(self, dirname: str):
        """
        the "." always refer to `.piperider` not the current working directory
        """
        if dirname is None or dirname.strip() == '':
            dirname = '.'
        if dirname.startswith('.'):
            return os.path.abspath(os.path.join(os.getcwd(), PIPERIDER_WORKSPACE_NAME, dirname))
        return os.path.abspath(dirname)

    def get_output_dir(self):
        if self.report_dir is None:
            return os.path.join(piperider_default_report_dir, 'outputs')
        return os.path.join(self.report_dir, 'outputs')

    def get_comparison_dir(self):
        if self.report_dir is None:
            return os.path.join(piperider_default_report_dir, 'comparisons')
        return os.path.join(self.report_dir, 'comparisons')

    def get_report_dir(self):
        return self.report_dir



import os
import sys


class PipeRiderError(Exception):
    """ Base class for piperider errors. """

    def __init__(self, *args, **kwargs):
        self.message = args[0] if len(args) else ''
        self.hint = kwargs.get('hint') or ''

    def __str__(self):
        return self.message

    hint = ''
    message = ''


class PipeRiderConfigError(PipeRiderError):
    def __init__(self, config_file):
        self.config_file = config_file
        pass

    message = "Piperider is not initialized."
    hint = "Please execute command 'piperider init' to move forward."


class PipeRiderConfigTypeError(PipeRiderError):
    def __init__(self, msg):
        self.message = msg

    hint = "Please check your input configuration in config.yml"


class PipeRiderCredentialFieldError(PipeRiderError):
    def __init__(self, field, message):
        self.message = message
        self.hint = f"Please check '{field}' in credentials.yml"


class DbtProfileBigQueryAuthWithTokenUnsupportedError(PipeRiderError):
    def __init__(self):
        self.message = "PipeRider haven't supported dbt BigQuery method with 'oauth-secrets' yet."

    hint = "Please reference the document of dbt BigQuery Profile for more details.\n" \
           "    ref: https://docs.getdbt.com/reference/warehouse-profiles/bigquery-profile#authentication-methods"


class PipeRiderNoProfilingResultError(PipeRiderError):
    def __init__(self, result_file):
        self.result_file = result_file
        pass

    message = "No profiling result is found."
    hint = "Please execute command 'piperider run' to move forward."


class PipeRiderInvalidDataSourceError(PipeRiderError):
    def __init__(self, name, config_file):
        self.name = name
        self.config_file = config_file
        self.message = f"Unsupported data source type '{name}'."
        self.hint = f"Please modify the config file '{config_file}'."


class PipeRiderConnectorError(PipeRiderError):
    def __init__(self, err_msg, datasource_name):
        self.message = err_msg
        self.hint = f'Please run \"pip install \'piperider[{datasource_name}]\'\" to get the {datasource_name} connector'


class PipeRiderTableConnectionError(PipeRiderError):
    def __init__(self, name, type_name):
        self.message = f'No available table found or no access permission found from \'{name}\' data source.'
        self.hint = f'Please verify your {type_name} data source with correct access permission.'


class PipeRiderDataBaseConnectionError(PipeRiderError):
    def __init__(self, name, type_name, db_path=None):
        self.message = f'No available database found from \'{name}\' data source.'
        if db_path:
            self.message += f' Cannot access the database file: \'{db_path}\''
        self.hint = f'Please verify your {type_name} data source with correct access permission.'


class PipeRiderDataBaseEncodingError(PipeRiderError):
    def __init__(self, file_path, type_name, current_encoding, support_encoding):
        if current_encoding is None:
            self.message = f'Unsupported file encoding format. ' \
                           f'Currently we only support \'{support_encoding}\' encoding format in {type_name} data source.'
            self.hint = ''
        else:
            self.message = f'Unsupported file encoding format \'{current_encoding}\'. ' \
                           f'Currently we only support \'{support_encoding}\' encoding format in {type_name} data source.'
            if sys.platform == 'nt':
                self.hint = 'Please use the following URL to convert file encoding.\n' \
                            '    https://www.freeformatter.com/convert-file-encoding.html'
            else:  # Linux or Mac platform
                self.hint = f'Please use the following command to convert file encoding.\n' \
                            f'    \'iconv -f {current_encoding} -t {support_encoding} "{file_path}" > "utf8-{os.path.basename(file_path)}"\''


class DbtError(PipeRiderError):
    type = 'dbt'


class DbtProjectNotFoundError(DbtError):
    def __init__(self, dbt_project_path):
        self.dbt_project_path = dbt_project_path
        self.message = f"Cannot find dbt project at '{dbt_project_path}'."

    hint = "Please make sure the dbt project is in the same directory with the piperider config file."


class DbtProjectInvalidError(DbtError):
    def __init__(self, dbt_project_path, err_msg):
        self.dbt_project_path = dbt_project_path
        self.message = f"Failed to load dbt project '{dbt_project_path}'. \n Reason: {err_msg}"

    hint = "Please use 'dbt debug' to verify the dbt configuration."


class DbtProfileNotFoundError(DbtError):
    def __init__(self, dbt_profile_path):
        self.dbt_project_path = dbt_profile_path
        self.message = f"Cannot find dbt profiles at '{dbt_profile_path}'."

    hint = "Please use 'dbt init' to initiate the dbt profiles."


class DbtProfileInvalidError(DbtError):
    def __init__(self, dbt_profile_path, err_msg):
        self.dbt_profile_path = dbt_profile_path
        self.message = f"Failed to load dbt profile '{dbt_profile_path}'. \nReason: {err_msg}"

    hint = "Please use 'dbt debug' to verify the dbt configuration."


class DbtInvocationError(DbtError):
    def __init__(self, message):
        self.message = message

    hint = "Please reference dbt documentation for more information. ref: https://docs.getdbt.com/reference/exit-codes"


class DbtCommandNotFoundError(PipeRiderError):
    def __init__(self):
        self.message = "dbt command not found."
        self.hint = 'Please run \'pip install dbt-core\' to install dbt Core.'


class DbtAdapterCommandNotFoundError(PipeRiderError):
    def __init__(self, datasource_type):
        self.message = "dbt command not found."
        self.hint = f'Please run \'pip install dbt-{datasource_type}\' to install dbt Core and the adapter.'


class AssertionError(PipeRiderError):
    def __init__(self, error_msg):
        self.message = error_msg
        pass


class IllegalStateAssertionError(PipeRiderError):
    def __init__(self, error_msg):
        self.message = error_msg
        pass


class AwsCredentialsError(PipeRiderError):
    def __init__(self, error_msg, type='redshift'):
        self.type = type
        self.message = error_msg
        self.hint = 'Please configure the AWS credentials by command "aws configure".\n  Or setup the environment variables "AWS_ACCESS_KEY_ID" & "AWS_SECRET_ACCESS_KEY".'
        pass


class AwsUnExistedS3Bucket(PipeRiderError):
    def __init__(self, bucket):
        self.message = f'S3 bucket "{bucket}" does not exist'


class RecipeException(PipeRiderError):
    def __init__(self, message):
        self.message = message
        self.hint = 'Please use "git stash" to stash your changes or "git commit" to permanently save changes to current branch.'


class RecipeConfigException(PipeRiderError):
    def __init__(self, message, hint=None):
        self.message = message
        self.hint = hint



import os
from pathlib import Path

from jsonschema.exceptions import ValidationError
from rich.console import Console
from rich.syntax import Syntax

from piperider_cli.dbtutil import load_dbt_project
from piperider_cli.recipes import DEFAULT_RECIPE_PATH, RecipeConfiguration, RecipeModel, RecipeDbtField, \
    RecipePiperiderField
from piperider_cli.recipes.utils import git_branch

console = Console()


def _read_dbt_project_file(dbt_project_path):
    if dbt_project_path is None:
        return None
    try:
        dbt_project = load_dbt_project(dbt_project_path)
    except Exception as e:
        console.print(f'[[bold yellow]Skip[/bold yellow]] dbt project: {e}')
        return None
    return dbt_project


def _create_base_recipe(dbt_project_path=None, options: dict = None) -> RecipeModel:
    """
    Create the base recipe
    """
    base = RecipeModel()

    if git_branch() is not None:
        base.branch = 'main'

    dbt_project = _read_dbt_project_file(dbt_project_path)
    if dbt_project:
        base.dbt = RecipeDbtField({
            'commands': [
                'dbt deps',
                'dbt build'
            ]
        })

    base.piperider = RecipePiperiderField({
        'command': 'piperider run'
    })
    return base


def _create_target_recipe(dbt_project_path=None, options: dict = None) -> RecipeModel:
    """
    Create the target recipe
    """
    target = RecipeModel()

    current_branch = git_branch()
    if current_branch is not None and current_branch != 'main':
        target.branch = current_branch

    dbt_project = _read_dbt_project_file(dbt_project_path)
    if dbt_project:
        target.dbt = RecipeDbtField({
            'commands': [
                'dbt deps',
                'dbt build'
            ]
        })

    target.piperider = RecipePiperiderField({
        'command': 'piperider run'
    })
    return target


def generate_default_recipe(overwrite_existing: bool = False,
                            dbt_project_path=None, options: dict = None):
    """
    Generate the default recipe
    """
    recipe_path = DEFAULT_RECIPE_PATH
    if overwrite_existing is False and os.path.exists(recipe_path):
        console.print('[bold green]Piperider default recipe already exist[/bold green]')
        return 0
    base = _create_base_recipe(dbt_project_path)
    target = _create_target_recipe(dbt_project_path)
    recipe = RecipeConfiguration(base=base, target=target)

    try:
        Path(recipe_path).parent.mkdir(parents=True, exist_ok=True)
        recipe.dump(recipe_path)
    except ValidationError as e:
        console.print(f'[[bold red]Error[/bold red]] Recipe syntax error: {e}')
        return 1
    except Exception as e:
        console.print(f'[[bold red]Error[/bold red]] {e}')
        return 1

    return 0


def show_recipe_content(recipe_path=DEFAULT_RECIPE_PATH):
    """
    Display the recipe content
    """
    if not os.path.exists(recipe_path):
        # Skip if the recipe does not exist
        return

    with open(recipe_path, 'r') as f:
        console.rule(f'Recipe: {os.path.relpath(recipe_path)}')
        yaml_output = Syntax(f.read(), 'yaml', theme='monokai', line_numbers=True)
        console.print(yaml_output)
        console.rule('End of Recipe')


if __name__ == '__main__':
    generate_default_recipe(overwrite_existing=True)



from piperider_cli.assertion_engine import AssertionContext, AssertionResult
from piperider_cli.assertion_engine.assertion import ValidationResult
from piperider_cli.assertion_engine.types.base import BaseAssertionType


class AssertRowCountInRange(BaseAssertionType):

    def name(self):
        return "assert_row_count_in_range"

    def execute(self, context: AssertionContext):
        return assert_row_count_in_range(context)

    def validate(self, context: AssertionContext) -> ValidationResult:
        return ValidationResult(context).require('count').require_int_pair('count')


class AssertRowCount(BaseAssertionType):
    def name(self):
        return "assert_row_count"

    def execute(self, context: AssertionContext):
        return assert_row_count(context)

    def validate(self, context: AssertionContext) -> ValidationResult:
        results = ValidationResult(context) \
            .require_one_of_parameters(['min', 'max']) \
            .int_if_present('min') \
            .int_if_present('max')

        if results.errors:
            return results

        if context.asserts.get('min') is not None and context.asserts.get('max') is not None:
            if context.asserts.get('min') > context.asserts.get('max'):
                results.errors.append('The max value should be greater than or equal to the min value.')

        return results


def assert_row_count(context: AssertionContext) -> AssertionResult:
    table = context.table
    column = context.column
    metrics = context.profiler_result

    table_metrics = metrics.get('tables', {}).get(table)
    if not table_metrics:
        return context.result.fail_with_metric_not_found_error(table, column)

    # Get the metric for the current table
    row_count = table_metrics.get('row_count')
    context.result.actual = row_count

    min = context.asserts.get('min', 0)
    max = context.asserts.get('max')

    if context.asserts.get('min') is None and context.asserts.get('max') is None:
        return context.result.fail_with_assertion_error('Expect a min or max value.')
    if not isinstance(min, int):
        return context.result.fail_with_assertion_error('The min value should be an integer.')
    if min < 0:
        return context.result.fail_with_assertion_error('The min value should be greater than or equal to 0.')

    # Only provide min
    if max is None:
        if min <= row_count:
            return context.result.success()
    else:
        if not isinstance(max, int):
            return context.result.fail_with_assertion_error('The max value should be an integer.')
        if max < min:
            return context.result.fail_with_assertion_error(
                'The max value should be greater than or equal to the min value.')
        if min <= row_count <= max:
            return context.result.success()
    return context.result.fail()


def assert_row_count_in_range(context: AssertionContext) -> AssertionResult:
    table = context.table
    column = context.column
    metrics = context.profiler_result

    table_metrics = metrics.get('tables', {}).get(table)
    if not table_metrics:
        return context.result.fail_with_metric_not_found_error(table, column)

    # Get the metric for the current table
    row_count = table_metrics.get('row_count')
    context.result.actual = row_count

    between_criteria = context.asserts.get('count', [])

    if len(between_criteria) != 2:
        return context.result.fail_with_assertion_error('Expect a range [min_value, max_value].')

    valid_type = isinstance(between_criteria[0], int) and isinstance(between_criteria[1], int)
    if not valid_type:
        return context.result.fail_with_assertion_error('The range should be integers.')

    if between_criteria[0] > between_criteria[1]:
        return context.result.fail_with_assertion_error(
            'The minimum value of the range should be the first number.')

    if between_criteria[0] <= row_count <= between_criteria[1]:
        return context.result.success()

    return context.result.fail()



import os
from unittest import TestCase

import piperider_cli.dbtutil as dbtutil
from piperider_cli.datasource.sqlite import SqliteDataSource
from piperider_cli.profiler import Profiler
from tests.common import create_table


class TestRunner(TestCase):

    def setUp(self):
        self.dbt_state_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "mock_dbt_data")

    def test_get_dbt_state_candidate(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=None,
                                                                          dbt_run_results=None,
                                                                          tag=None))
        self.assertEqual(len(tables), 6)
        self.assertEqual(tables[0].get('name'), 'PRICE_PRESENT')
        self.assertEqual(tables[0].get('schema'), 'PUBLIC')
        self.assertEqual(tables[0].get('alias'), 'PRICE_PRESENT')
        self.assertEqual(tables[1].get('name'), 'event_filtered')
        self.assertEqual(tables[2].get('name'), 'session')
        self.assertEqual(tables[3].get('name'), 'event_enhanced')
        self.assertEqual(tables[4].get('name'), 'user')
        self.assertEqual(tables[5].get('name'), 'project')

    def test_get_dbt_state_candidate_view_profile(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=True,
                                                                          dbt_resources=None,
                                                                          dbt_run_results=None,
                                                                          tag=None))
        self.assertEqual(len(tables), 7)
        self.assertEqual(tables[0].get('name'), 'PRICE_PRESENT')
        self.assertEqual(tables[1].get('name'), 'event_filtered')
        self.assertEqual(tables[2].get('name'), 'stg_event')
        self.assertEqual(tables[3].get('name'), 'session')
        self.assertEqual(tables[4].get('name'), 'event_enhanced')
        self.assertEqual(tables[5].get('name'), 'user')
        self.assertEqual(tables[6].get('name'), 'project')

    def test_get_dbt_state_candidate_run_results(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=None,
                                                                          dbt_run_results=True,
                                                                          tag=None))
        self.assertEqual(len(tables), 1)
        self.assertEqual(tables[0].get('name'), 'PRICE_PRESENT')
        self.assertEqual(tables[0].get('schema'), 'PUBLIC')
        self.assertEqual(tables[0].get('alias'), 'PRICE_PRESENT')

    def test_get_dbt_state_candidate_resources_only_models(self):
        models = ['infusetude.amplitude.session',
                  'infusetude.amplitude.user']
        resources = dict(models=models, metrics=[])
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=resources,
                                                                          dbt_run_results=None,
                                                                          tag=None))
        self.assertEqual(len(tables), 2)
        self.assertEqual(tables[0].get('name'), 'session')
        self.assertEqual(tables[1].get('name'), 'user')

    def test_get_dbt_state_candidate_resources_only_seeds(self):
        models = ['infusetude.project_block',
                  'infusetude.user_block']
        resources = dict(models=models, metrics=[])
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=resources,
                                                                          dbt_run_results=None,
                                                                          tag=None))
        self.assertEqual(len(tables), 0)

    def test_get_dbt_state_candidate_resources_only_metrics(self):
        metrics = ['metric.infusetude.active_projects',
                   'metric.infusetude.active_projects_per_user',
                   'metric.infusetude.active_users',
                   'metric.infusetude.total_events',
                   'metric.infusetude.total_sessions',
                   'metric.infusetude.total_users']
        resources = dict(models=[], metrics=metrics)
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=resources,
                                                                          dbt_run_results=None,
                                                                          tag=None))
        self.assertEqual(len(tables), 0)

    def test_get_dbt_state_candidate_resources_not_found_or_empty(self):
        models = ['infusetude.amplitude.notfound']
        resources = dict(models=models, metrics=[])
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=resources,
                                                                          dbt_run_results=None,
                                                                          tag=None))
        self.assertEqual(len(tables), 0)

        resources = dict(models=[], metrics=[])
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=resources,
                                                                          dbt_run_results=None,
                                                                          tag=None))
        self.assertEqual(len(tables), 0)

    def test_get_dbt_state_candidate_resources_without_view_profile(self):
        models = ['infusetude.amplitude.stg_event',
                  'infusetude.amplitude.session',
                  'infusetude.amplitude.user']
        resources = dict(models=models, metrics=[])
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=resources,
                                                                          dbt_run_results=None,
                                                                          tag=None))
        self.assertEqual(len(tables), 3)
        self.assertEqual(tables[0].get('name'), 'stg_event')
        self.assertEqual(tables[1].get('name'), 'session')
        self.assertEqual(tables[2].get('name'), 'user')

    def test_get_dbt_state_candidate_tag(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=None,
                                                                          dbt_run_results=None,
                                                                          tag='test'))
        self.assertEqual(len(tables), 2)
        self.assertEqual(tables[0].get('name'), 'stg_event')
        self.assertEqual(tables[1].get('name'), 'session')

    def test_get_dbt_state_candidate_tag_and_view(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=True,
                                                                          dbt_resources=None,
                                                                          dbt_run_results=None,
                                                                          tag='test'))
        self.assertEqual(len(tables), 2)
        self.assertEqual(tables[0].get('name'), 'stg_event')
        self.assertEqual(tables[1].get('name'), 'session')

    def test_get_dbt_state_candidate_tag_and_run_results(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=None,
                                                                          dbt_run_results=True,
                                                                          tag='piperider'))
        self.assertEqual(len(tables), 1)
        self.assertEqual(tables[0].get('name'), 'PRICE_PRESENT')

    def test_get_dbt_state_candidate_tag_and_run_results_emptyset(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=None,
                                                                          dbt_run_results=True,
                                                                          tag='test'))
        self.assertEqual(len(tables), 0)

    def test_get_dbt_state_candidate_view_profile_tag_and_run_results(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=True,
                                                                          dbt_resources=None,
                                                                          dbt_run_results=True,
                                                                          tag='piperider'))
        self.assertEqual(len(tables), 1)
        self.assertEqual(tables[0].get('name'), 'PRICE_PRESENT')

    def test_get_dbt_state_tests_result(self):
        results = dbtutil.get_dbt_state_tests_result(self.dbt_state_dir)

        table_names = [r.get('table') for r in results]
        column_names = [r.get('column') for r in results]

        self.assertIn('PRICE_PRESENT', table_names)
        self.assertIn('MA60', column_names)

        status = None
        for r in results:
            if r.get('table') == 'PRICE_PRESENT' and r.get('column') == 'MA60':
                status = r.get('status')
                break
        self.assertEqual('failed', status)

    def test_dbt_append_descriptions(self):
        ds = SqliteDataSource("test")
        engine = ds.get_engine_by_database()
        profiler = Profiler(ds)

        data1 = [
            ("symbol", "open", "ma60", "ma20"),
            ("AAA", 1, 2, 3),
            ("BBB", 4, 5, 6),
        ]
        data2 = [
            ("symbol", "open", "ma60", "ma20"),
            ("CCC", 1, 2, 3),
            ("DDD", 4, 5, 6),
        ]
        create_table(engine, "PRICE", data1)
        create_table(engine, "PRICE_PRESENT", data2)
        profile_results = profiler.profile()
        dbtutil.append_descriptions(profile_results, self.dbt_state_dir)

        self.assertEqual('The price data to today - via DBT',
                         profile_results['tables']['PRICE_PRESENT']['description'])
        self.assertEqual('The symbol name - via DBT',
                         profile_results['tables']['PRICE_PRESENT']['columns']['symbol']['description'])

    def test_get_dbt_state_metrics_only_tag(self):
        metrics = dbtutil.get_dbt_state_metrics(self.dbt_state_dir, 'piperider', None)

        self.assertEqual(len(metrics), 3)

        metrics = dbtutil.get_dbt_state_metrics(self.dbt_state_dir, 'abc', None)

        self.assertEqual(len(metrics), 1)
        self.assertEqual(metrics[0].name, 'active_projects_per_user')

    def test_get_dbt_state_metrics_only_resources(self):
        metrics = ['metric.infusetude.active_projects',
                   'metric.infusetude.active_projects_per_user',
                   'metric.infusetude.active_users',
                   'metric.infusetude.total_events',
                   'metric.infusetude.total_users']
        resources = dict(models=[], metrics=metrics)
        metrics = dbtutil.get_dbt_state_metrics(self.dbt_state_dir, 'piperider', resources)
        self.assertEqual(len(metrics), 5)

        metrics = dbtutil.get_dbt_state_metrics(self.dbt_state_dir, 'abc', resources)
        self.assertEqual(len(metrics), 5)
        self.assertEqual(metrics[0].name, 'total_events')



import inquirer

from piperider_cli.error import PipeRiderConnectorError, PipeRiderCredentialFieldError
from . import DataSource
from .field import TextField, PasswordField, ListField, PathField, DataSourceField

AUTH_METHOD_PASSWORD = 'password'
AUTH_METHOD_KEYPAIR = 'keypair'
AUTH_METHOD_SSO = 'sso'


class SnowflakeDataSource(DataSource):
    def __init__(self, name, **kwargs):
        super().__init__(name, 'snowflake', **kwargs)
        # self.fields = ["account", "user", "password", "database", "warehouse", "schema"]
        self.fields = [
            TextField('account', description='Account'),
            TextField('user', description='Username'),
            ListField('_method', description='Authentication Methods',
                      default=[AUTH_METHOD_PASSWORD, AUTH_METHOD_KEYPAIR, AUTH_METHOD_SSO]),

            PasswordField('password', description='Password',
                          ignore=lambda answers: answers.get('_method') != AUTH_METHOD_PASSWORD),
            PathField('private_key_path', description='Private key path',
                      ignore=lambda answers: answers.get('_method') != AUTH_METHOD_KEYPAIR),
            PasswordField('private_key_passphrase', description='Passphrase for the private key', optional=True,
                          ignore=lambda answers: answers.get('_method') != AUTH_METHOD_KEYPAIR),
            TextField('authenticator', description='Authenticator (\'externalbrowser\' or a valid Okta URL)',
                      default=lambda answers: 'externalbrowser' if answers.get('_method') == AUTH_METHOD_SSO else None,
                      ignore=lambda answers: answers.get('_method') != AUTH_METHOD_SSO),
            TextField('role', description='Role', optional=True),
            TextField('database', description='Database'),
            TextField('warehouse', description='Warehouse'),
            TextField('schema', default='PUBLIC', description='Schema'),
        ]
        self._connect_timeout = 5

    def validate(self):
        if self.type_name != 'snowflake':
            raise ValueError('type name should be snowflake')
        return self._validate_required_fields()

    def to_database_url(self, database):
        credential = self.credential
        account = credential.get('account')
        password = credential.get('password')
        user = credential.get('user')
        if database is None:
            database = credential.get('database')
        schema = credential.get('schema')
        warehouse = credential.get('warehouse')
        role = credential.get('role')
        authenticator = credential.get('authenticator')
        from snowflake.sqlalchemy.snowdialect import SnowflakeDialect
        from snowflake.sqlalchemy import URL

        SnowflakeDialect.supports_statement_cache = True
        db_parameters = {
            "account": account,
            "user": user,
            "database": database,
            "schema": schema,
            "warehouse": warehouse
        }

        if password:
            db_parameters["password"] = password

        if role:
            db_parameters["role"] = role

        if authenticator:
            db_parameters["authenticator"] = authenticator

        return URL(**db_parameters)

    def engine_args(self):
        return dict(connect_args={
            'connect_timeout': self._connect_timeout,
            'private_key': self._get_private_key(),
        })

    def _get_private_key(self):
        private_key_path = self.credential.get('private_key_path')
        private_key_passphrase = self.credential.get('private_key_passphrase')

        if private_key_path is None:
            return None

        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives import serialization

        if private_key_passphrase:
            encoded_passphase = private_key_passphrase.encode()
        else:
            encoded_passphase = None

        with open(private_key_path, "rb") as key:
            p_key = serialization.load_pem_private_key(
                key.read(),
                password=encoded_passphase,
                backend=default_backend()
            )

        return p_key.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption())

    def verify_connector(self):
        try:
            import snowflake.connector
            import snowflake.sqlalchemy
            # do nothing when everything is ok
            return None
        except Exception as e:
            return PipeRiderConnectorError(str(e), 'snowflake')

    def _get_display_description(self):
        cred = self.credential
        return f"type={self.type_name}, database={cred.get('database')}, schema={cred.get('schema')}"

    def get_database(self):
        return self.credential.get('database')

    def get_schema(self):
        return self.credential.get('schema')



import os
from unittest import TestCase

from piperider_cli.runner import _filter_subject, get_dbt_profile_subjects
from piperider_cli.profiler import ProfileSubject


class TestRunner(TestCase):

    def setUp(self):
        self.dbt_state_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "mock_dbt_data")

    def test_filter_subject(self):
        self.assertTrue(_filter_subject('a', None, None))
        self.assertTrue(_filter_subject('a', ['a', 'b'], None))
        self.assertFalse(_filter_subject('a', [], None))
        self.assertFalse(_filter_subject('a', None, ['a', 'b']))
        self.assertTrue(_filter_subject('a', None, []))
        self.assertFalse(_filter_subject('a', ['a'], ['a']))
        self.assertTrue(_filter_subject('a', ['a'], ['b']))
        self.assertTrue(_filter_subject('a', ['A'], ['B']))

    def test_dbt_profile_subjects_tags_and_config(self):
        options = dict(
            view_profile=False,
            dbt_resources=None,
            dbt_run_results=None,
            tag='test'
        )

        def filter_fn(subject: ProfileSubject):
            return _filter_subject(subject.name, ['project'], None)

        subjects = get_dbt_profile_subjects(self.dbt_state_dir, options, filter_fn)

        self.assertEqual(2, len(subjects))
        self.assertEqual('stg_event', subjects[0].name)
        self.assertEqual('session', subjects[1].name)

    def test_dbt_profile_subjects_dbt_resources_and_config(self):
        options = dict(
            view_profile=None,
            dbt_resources=dict(
                models=['infusetude.amplitude.project'],
                metrics=[]
            ),
            dbt_run_results=None,
            tag=None
        )

        def filter_fn(subject: ProfileSubject):
            return _filter_subject(subject.name, None, ['project'])

        subjects = get_dbt_profile_subjects(self.dbt_state_dir, options, filter_fn)

        self.assertEqual(1, len(subjects))
        self.assertEqual('project', subjects[0].name)

    def test_dbt_profile_subjects_config_excludes(self):
        options = dict(
            view_profile=None,
            dbt_resources=None,
            dbt_run_results=None,
            tag=None
        )

        def filter_fn(subject: ProfileSubject):
            return _filter_subject(subject.name, None, ['session'])

        subjects = get_dbt_profile_subjects(self.dbt_state_dir, options, filter_fn)

        self.assertEqual(5, len(subjects))
        self.assertNotIn('session', [s.name for s in subjects])

    def test_dbt_profile_subjects_config_includes(self):
        options = dict(
            view_profile=None,
            dbt_resources=None,
            dbt_run_results=None,
            tag=None
        )

        def filter_fn(subject: ProfileSubject):
            return _filter_subject(subject.name, ['user'], None)

        subjects = get_dbt_profile_subjects(self.dbt_state_dir, options, filter_fn)

        self.assertEqual(1, len(subjects))
        self.assertEqual('user', subjects[0].name)

    def test_dbt_profile_subjects_config_view_includes(self):
        options = dict(
            view_profile=True,
            dbt_resources=None,
            dbt_run_results=None,
            tag=None
        )

        def filter_fn(subject: ProfileSubject):
            return _filter_subject(subject.name, ['stg_event'], None)

        subjects = get_dbt_profile_subjects(self.dbt_state_dir, options, filter_fn)

        self.assertEqual(1, len(subjects))
        self.assertEqual('stg_event', subjects[0].name)

    def test_dbt_profile_subjects_config_no_view_includes(self):
        options = dict(
            view_profile=False,
            dbt_resources=None,
            dbt_run_results=None,
            tag=None
        )

        def filter_fn(subject: ProfileSubject):
            return _filter_subject(subject.name, ['stg_event', 'user'], None)

        subjects = get_dbt_profile_subjects(self.dbt_state_dir, options, filter_fn)

        self.assertEqual(1, len(subjects))
        self.assertEqual('user', subjects[0].name)



import os
import shlex
import subprocess
from subprocess import Popen, TimeoutExpired
from typing import Dict

from piperider_cli.error import RecipeException


def _execute_command(command_line, env: Dict = None):
    cmd = shlex.split(command_line)
    proc = Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env or os.environ.copy())
    try:
        outs, errs = proc.communicate(timeout=15)
    except TimeoutExpired:
        proc.kill()
        outs, errs = proc.communicate()

    if outs is not None:
        outs = outs.decode().strip()
    if errs is not None:
        errs = errs.decode().strip()
    return outs, errs, proc.returncode


def execute_command_without_capture_output(command_line, env: Dict = None):
    cmd = shlex.split(command_line)
    proc = Popen(cmd, env=env or os.environ.copy())
    try:
        proc.communicate()
    except TimeoutExpired:
        proc.kill()
        proc.communicate()

    return proc.returncode


def git_branch():
    outs, errs, exit_code = _execute_command("git rev-parse --abbrev-ref HEAD")
    if exit_code != 0:
        return None

    return outs


def git_switch_to(branch_name):
    outs, errs, exit_code = _execute_command(f"git switch {branch_name}")
    if exit_code != 0:
        raise RecipeException(errs)


def execute_command(command_line, envs: Dict):
    exit_code = execute_command_without_capture_output(command_line, envs)
    return exit_code


def ensure_git_ready():
    outs, errs, exit_code = _execute_command("git --version")
    if exit_code != 0:
        raise RecipeException(errs)

    if "version" not in outs:
        raise RecipeException("Unknown response from git --version")

    outs, errs, exit_code = _execute_command("git status --porcelain")
    if exit_code != 0 and "not a git repository" in errs:
        raise RecipeException("The working directory is not a git repository.")

    dirty_list = [x for x in outs.split("\n") if not x.strip().startswith("??")]
    dirty_list = [x for x in dirty_list if x]
    if len(dirty_list) != 0:
        raise RecipeException("Working directory is dirty. Stop to run the recipe")


def check_dbt_command():
    outs, errs, exit_code = _execute_command("dbt --version")
    if exit_code != 0:
        raise RecipeException(errs)


if __name__ == '__main__':
    ensure_git_ready()



import os
from abc import ABCMeta, abstractmethod
from typing import List, Callable

import inquirer
from rich.console import Console
from rich.prompt import Prompt


def _default_validate_func(answer, current) -> bool:
    if current is None:
        return False
    if isinstance(current, str):
        return current.strip() != ''
    return True


class DataSourceField(metaclass=ABCMeta):

    def __init__(self, name, type, value=None, default=None, description=None, validate=_default_validate_func,
                 optional=False, ignore=False):
        self.name = name
        self.type = type
        self.default = default
        self.description = description if not optional else f'{description} (optional)'
        self.value = value
        self.optional = optional
        self.ignore = ignore
        self.callback: Callable = None
        if not isinstance(ignore, bool) and not isinstance(ignore, Callable):
            raise ValueError('ignore must be a callable function or a boolean type')
        if optional is True:
            self.validate = True
        else:
            self.validate = validate

    def get(self):
        return self.value

    def set(self, value):
        self.value = value

    @abstractmethod
    def question(self):
        raise NotImplementedError

    def choose_by_rich(self):
        choices = self.default
        console = Console()
        console.print(f'[[yellow]?[/yellow]] Please choose one of the following {self.description}:')
        i = 0
        for v in choices:
            console.print(f'  [green]{i + 1}[/green]: {v}')
            i += 1

        while True:
            try:
                type_idx = Prompt.ask('[[yellow]?[/yellow]] Select a number')
                type_idx = int(type_idx)
            except Exception:
                type_idx = 0
            if type_idx > len(choices) or type_idx < 1:
                console.print('    [[red]Error[/red]] Input is not a valid index value. Please try again.')
            else:
                answer = choices[type_idx - 1]
                break
        return answer if answer is not None else ''

    def question_by_rich(self, answers):
        default = self.default
        password = False
        is_path = False

        ignore = self.ignore
        if isinstance(ignore, bool) and ignore is True or isinstance(ignore, Callable) and ignore(answers):
            return None
        if isinstance(default, Callable):
            default = default(answers)

        if self.type == 'password':
            password = True
        elif self.type == 'path':
            is_path = True
        elif self.type == 'number':
            default = str(default)
        elif self.type == 'list':
            return self.choose_by_rich()

        console = Console()
        while True:
            answer = Prompt.ask(f'[[yellow]?[/yellow]] {self.description}', password=password, default=default)
            if is_path and answer:
                answer = os.path.expanduser(answer)
            if self.validate is None \
                or (isinstance(self.validate, bool) and self.validate is True) \
                or self.validate(None,
                                 answer):
                # Passed validation
                if is_path is True and os.path.exists(answer) is False:
                    console.print('    [[red]Error[/red]] Input path does not exist.')
                    continue
                break
            else:
                if answer is None:
                    console.print('    [[red]Error[/red]] Input value is required. Please try again.')
                else:
                    console.print(f'    [[red]Error[/red]] Invalid input: {answer}')
        return answer if answer is not None else ''


class TextField(DataSourceField):
    def __init__(self, name, value=None, default=None, description=None, validate=_default_validate_func,
                 optional=False, ignore=False):
        super().__init__(name, "text", value, default, description, validate, optional, ignore)

    def question(self):
        return inquirer.Text(self.name, message=self.description, default=self.default, validate=self.validate,
                             ignore=self.ignore)


class PathField(DataSourceField):
    def __init__(self, name, value=None, default=None, description=None, validate=_default_validate_func,
                 optional=False, ignore=False):
        super().__init__(name, "path", value, default, description, validate, optional, ignore)

    def question(self):
        return inquirer.Path(self.name, message=self.description, default=self.default, normalize_to_absolute_path=True,
                             exists=True, ignore=self.ignore)


class NumberField(DataSourceField):
    def __init__(self, name, value=None, default=None, description=None, validate=_default_validate_func,
                 optional=False, ignore=False):
        def _is_numeric_func(answer, current) -> bool:
            return current.strip().isnumeric()

        def _is_numeric_or_empty_func(answer, current) -> bool:
            return current.strip().isnumeric() or current == ''

        if value and not (isinstance(value, int) or isinstance(value, float)):
            raise ValueError("Value must be a number")
        if default and not (isinstance(default, int) or isinstance(default, float)):
            raise ValueError("Default value must be a number")
        if optional:
            validate = _is_numeric_or_empty_func
        else:
            validate = _is_numeric_func
        super().__init__(name, "number", value, default, description, validate, optional, ignore)

    def question(self):
        return inquirer.Text(self.name, message=self.description, default=str(self.default), validate=self.validate,
                             ignore=self.ignore)


class PasswordField(DataSourceField):
    def __init__(self, name, value=None, default=None, description=None, validate=_default_validate_func,
                 optional=False, ignore=False):
        super().__init__(name, "password", value, default, description, validate, optional, ignore)

    def question(self):
        return inquirer.Password(self.name, message=self.description, default=self.default, validate=self.validate,
                                 ignore=self.ignore)


class ListField(DataSourceField):
    def __init__(self, name, value=None, default=None, description=None, validate=_default_validate_func,
                 optional=False, ignore=False):
        if not isinstance(default, List):
            raise ValueError("[ListField] default must be a list. Please provide the choices in default.")
        super().__init__(name, "list", value, default, description, validate, optional, ignore)

    def question(self):
        return inquirer.List(self.name, message=self.description, choices=self.default, validate=self.validate,
                             ignore=self.ignore)



from piperider_cli.assertion_engine import AssertionContext, ValidationResult, AssertionResult
from piperider_cli.assertion_engine.types.assert_column_misc import \
    AssertColumnNotNull, \
    AssertColumnNull, \
    AssertColumnUnique, \
    AssertColumnExist, \
    AssertColumnValue
from piperider_cli.assertion_engine.types.assert_column_ranges import \
    AssertColumnMinInRange, \
    AssertColumnMaxInRange, \
    AssertColumnInRange
from piperider_cli.assertion_engine.types.assert_column_types import \
    AssertColumnSchemaType, \
    AssertColumnType, \
    AssertColumnInTypes
from piperider_cli.assertion_engine.types.assert_metrics import AssertMetric
from piperider_cli.assertion_engine.types.assert_rows import \
    AssertRowCountInRange, \
    AssertRowCount
from piperider_cli.assertion_engine.types.base import BaseAssertionType

custom_registry = {}


class _NotFoundAssertion(BaseAssertionType):

    def __init__(self, assertion_name):
        self.assertion_name = assertion_name

    def name(self):
        return self.assertion_name

    def execute(self, context: AssertionContext) -> AssertionResult:
        raise BaseException('this is not-found-assertion, cannot be used')

    def validate(self, context: AssertionContext) -> ValidationResult:
        result = ValidationResult(context)
        result.errors.append(f"cannot find an assertion by name '{context.name}'")
        return result


def register_assertion_function(typename: BaseAssertionType):
    instance: BaseAssertionType = typename()
    if instance.name() in custom_registry:
        raise ValueError(f'{instance.name()} has been registered by {custom_registry.get(instance.name())}')
    custom_registry[instance.name()] = instance


def get_assertion(name: str, metric: str) -> BaseAssertionType:
    if metric:
        return AssertMetric()

    if name not in custom_registry:
        return _NotFoundAssertion(name)
    return custom_registry[name]


register_assertion_function(AssertRowCountInRange)
register_assertion_function(AssertRowCount)

register_assertion_function(AssertColumnSchemaType)
register_assertion_function(AssertColumnType)
register_assertion_function(AssertColumnInTypes)

register_assertion_function(AssertColumnMinInRange)
register_assertion_function(AssertColumnMaxInRange)
register_assertion_function(AssertColumnInRange)

register_assertion_function(AssertColumnNotNull)
register_assertion_function(AssertColumnNull)
register_assertion_function(AssertColumnUnique)
register_assertion_function(AssertColumnExist)
register_assertion_function(AssertColumnValue)

if __name__ == '__main__':
    pass



#!/usr/bin/env python
import os
from distutils.core import setup

from setuptools import find_packages  # type: ignore


def _get_version():
    version_file = os.path.normpath(os.path.join(os.path.dirname(__file__), 'piperider_cli', 'VERSION'))
    with open(version_file) as fh:
        version = fh.read().strip()
        return version


duckdb_require_packages = [
    'duckdb>=0.4.0',
    'duckdb-engine>=0.6.1',
    'chardet>=5.0.0',
]

setup(name='piperider',
      version=_get_version(),
      description='PiperRider CLI',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      author='InfuseAI Dev Team',
      author_email='dev@infuseai.io',
      url='https://github.com/InfuseAI/piperider',
      entry_points={
          'console_scripts': ['piperider = piperider_cli.cli:cli']
      },
      python_requires=">=3.7",
      packages=find_packages(),
      install_requires=[
          'ruamel.yaml',
          'sqlalchemy>=1.3.18,<2.0',
          'sentry-sdk',
          'rich>=12.0.0',
          'click>=7.1',
          'python-dateutil',
          'inquirer>=2.10.0',
          'jinja2',
          "jsonschema>=3.2.0",
          'portalocker',
          'requests>=2.28.1',
          'requests_toolbelt>=0.9.1',
          'deepmerge',
      ],
      extras_require={
          'snowflake': [
              'snowflake-sqlalchemy'
          ],
          'postgres': [
              # you need a postgres for m1 to install psycopg2
              'psycopg2-binary'
          ],
          'bigquery': [
              'pyarrow<6.1.0,>=6.0.0',
              'pytz',
              'pybigquery',
              'google-cloud-bigquery-storage',

          ],
          'redshift': [
              'sqlalchemy-redshift',
              'redshift-connector',
              'psycopg2-binary',
              'boto3>=1.24.11',
          ],
          'duckdb': duckdb_require_packages,
          'csv': duckdb_require_packages,
          'parquet': duckdb_require_packages,
          'dev': [
              'pytest>=4.6',
              'pytest-flake8',
              'flake8==3.9.2',
              'pytest-mypy',
              'pytest-cov',
              'twine',
              'jsonschema',
          ],
      },
      project_urls={
          "Bug Tracker": "https://github.com/InfuseAI/piperider/issues",
      },
      classifiers=[
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
          "Development Status :: 4 - Beta"
      ],
      package_data={
          'piperider_cli': ['*.json', 'VERSION', 'SENTRY_DNS', 'data/**', 'profiler/schema.json',
                            'recipes/recipe_schema.json']
      })
